"""Tests the system end-to-end using real web calls."""
from functools import partial
import logging
import multiprocessing
from multiprocessing import Queue
import os
from pathlib import Path
import time
from typing import Any, cast
from unittest.mock import Mock

from _pytest.monkeypatch import MonkeyPatch
import pytest
from pytest_mock import MockerFixture

from bitfount import BitfountSchema, DataStructure
from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.csv_source import CSVSource
from bitfount.runners.config_schemas import ModellerConfig, PodConfig
from bitfount.runners.modeller_runner import setup_modeller_from_config
from bitfount.runners.pod_runner import setup_pod_from_config
from tests.integration import CONFIG_DIR
from tests.integration.bitfount_web_interactions import (
    WebdriverBitfountSession,
    get_bitfount_session,
    grant_proactive_access,
    oidc_flow,
)
from tests.integration.utils import (
    load_modeller_config,
    load_pod_config,
    password_from_modeller_config,
    password_from_pod_config,
    pod_start,
    run_modeller_process,
    tie_together_configs,
)
from tests.utils.helper import TABLE_NAME, backend_test, end_to_end_test

logger = logging.getLogger(__name__)

# Get timeouts/sleeps from envvars if possible
POD_STARTUP_SLEEP: int = int(os.getenv("E2E_TEST_POD_STARTUP_SLEEP", default=20))
MODELLER_STARTUP_SLEEP: int = int(
    os.getenv("E2E_TEST_MODELLER_STARTUP_SLEEP", default=5)
)
MODELLER_RUN_TIMEOUT: int = int(
    os.getenv("E2E_TEST_MODELLER_RUN_TIMEOUT", default=2.5 * 60)
)


@backend_test
@end_to_end_test
class TestCensusIncomeEndToEnd:
    """End-to-end tests using the Census Income dataset."""

    @pytest.fixture(autouse=True)
    def bitfount_env(self, monkeypatch: MonkeyPatch) -> None:
        """Set BITFOUNT_ENVIRONMENT env var to 'staging'."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")

    @pytest.fixture
    def token_dir(self, tmp_path: Path) -> Path:
        """Directory to store tokens."""
        return tmp_path / "bitfount_tokens"

    @pytest.fixture
    def modeller_config(self) -> ModellerConfig:
        """Modeller config."""
        return load_modeller_config(CONFIG_DIR / "modeller.yaml")

    @pytest.fixture
    def census_income_modeller_config(self) -> ModellerConfig:
        """Census income modeller config."""
        return load_modeller_config(
            CONFIG_DIR / "modeller_census_income_result_only.yaml"
        )

    @pytest.fixture
    def modeller_pswd(self, modeller_config: ModellerConfig) -> str:
        """Modeller password."""
        return password_from_modeller_config(modeller_config)

    @pytest.fixture
    def pod_1_config(self, census_income_data: Path) -> PodConfig:
        """Pod 1 config."""
        return load_pod_config(CONFIG_DIR / "pod_1.yaml", census_income_data)

    @pytest.fixture
    def census_income_pod_config(self, census_income_data: Path) -> PodConfig:
        """Census income pod config."""
        return load_pod_config(
            CONFIG_DIR / "pod_census_income.yaml", census_income_data
        )

    @pytest.fixture
    def pod_1_pswd(self, pod_1_config: PodConfig) -> str:
        """Pod 1 password."""
        return password_from_pod_config(pod_1_config)

    @pytest.fixture
    def pod_2_config(self, census_income_data: Path) -> PodConfig:
        """Pod 2 config."""
        return load_pod_config(CONFIG_DIR / "pod_2.yaml", census_income_data)

    @pytest.fixture
    def pod_2_pswd(self, pod_2_config: PodConfig) -> str:
        """Pod 2 password."""
        return password_from_pod_config(pod_2_config)

    @pytest.fixture
    def datasource(self, census_income_data: Path) -> BaseSource:
        """Census Income datasource."""
        return CSVSource(census_income_data, seed=100, ignore_cols=["fnlwgt"])

    @pytest.fixture
    def mock_modeller_oidc(self, mocker: MockerFixture) -> Any:
        """Mock out oidc auth flow."""
        oidc_webbrowser_import = (
            "bitfount.federated.transport.identity_verification"
            ".oidc.webbrowser.open_new_tab"
        )

        oidc_patcher = mocker.patch(oidc_webbrowser_import)
        return oidc_patcher

    @pytest.fixture
    def schema(self, datasource: BaseSource) -> BitfountSchema:
        """Census income schema."""
        return BitfountSchema(
            datasource,
            table_name=TABLE_NAME,
            force_stypes={
                TABLE_NAME: {
                    "categorical": [
                        "TARGET",
                        "workclass",
                        "marital-status",
                        "occupation",
                        "relationship",
                        "race",
                        "native-country",
                        "gender",
                        "education",
                    ]
                },
            },
        )

    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
    async def test_local_vs_pod(
        self,
        census_income_modeller_config: ModellerConfig,
        census_income_pod_config: PodConfig,
        datasource: BaseSource,
        mock_bitfount_session: Mock,
        mock_modeller_oidc: Mock,
        modeller_pswd: str,
        pod_1_pswd: str,
        schema: BitfountSchema,
        token_dir: Path,
    ) -> None:
        """Tests local vs remote results with oidc device code auth."""
        try:
            # Load configs and tie together
            census_income_modeller_config.pods.identifiers = [
                census_income_pod_config.pod_id
            ]
            # Mock BitfountSession
            # Override default return_value with side_effect
            # get_bitfount_session() one. side_effect takes precedence
            # over return_value.
            # If you ever change the order that the processes start in,
            # then be sure to change the order of these...
            mock_bitfount_session.side_effect = [
                # pod 1 bitfount session
                get_bitfount_session(
                    census_income_pod_config.username, pod_1_pswd, token_dir
                ),
                # modeller bitfount session
                get_bitfount_session(
                    census_income_modeller_config.modeller.username,
                    modeller_pswd,
                    token_dir,
                ),
            ]

            mock_modeller_oidc.side_effect = [
                partial(
                    oidc_flow,
                    username=census_income_modeller_config.modeller.username,
                    password=modeller_pswd,
                )
            ]
            # Load pods
            pod = setup_pod_from_config(census_income_pod_config)

            # Start pods
            logger.info("Spinning up pods... ")
            pod_process = multiprocessing.Process(
                target=pod_start, name="Pod_Runner", args=(pod,)
            )
            pod_process.start()

            time.sleep(POD_STARTUP_SLEEP)  # give the pods time to spin up

            assert pod_process.is_alive()

            logger.info("Pods loaded.")

            # Sort proactive access to each pod
            logger.info(
                f"Granting proactive access to pod {census_income_pod_config.pod_id}..."
            )
            grant_proactive_access(
                modeller_username=census_income_modeller_config.modeller.username,
                pod_id=census_income_pod_config.pod_id,
                role="general_modeller",
                # This is a WebdriverBitfountSession as we mocked it out
                pod_session=cast(WebdriverBitfountSession, pod._session),
            )

            # Load modeller
            # We do this after the pods have started, as we need the pods to have
            # published their public keys to the Hub.
            modeller, pod_identifiers = setup_modeller_from_config(
                census_income_modeller_config
            )

            queue: Queue = Queue()

            modeller_process = multiprocessing.Process(
                target=run_modeller_process,
                name="Modeller_Run",
                args=(
                    modeller,
                    pod_identifiers,
                    None,
                    queue,
                ),
            )

            # Start modeller
            logger.info("Starting Modeller...")
            modeller_process.start()
            time.sleep(MODELLER_STARTUP_SLEEP)  # give modeller time to spin up

            # Join modeller, wait for output for MODELLER_RUN_TIMEOUT seconds
            modeller_process.join(MODELLER_RUN_TIMEOUT)

            # Check modeller is done, pods still going
            logger.info("Modeller should be finished by this point.")
            assert not modeller_process.is_alive()
            assert (
                modeller_process.exitcode is not None and modeller_process.exitcode <= 0
            )

            pod_results = queue.get()[0]
            # Check pod_results exists
            assert pod_results
            # Ensure pods are still running
            assert pod_process.is_alive()

            # Stop all pods
            pod_process.terminate()
            pod_process.join()

            # Ensure shutdown
            assert pod_process.exitcode is not None and pod_process.exitcode <= 0

            schema.unfreeze()
            datastructure = DataStructure(table=pod.name, target="TARGET")
            hyperparameters = census_income_modeller_config.task.model.hyperparameters
            model_name = census_income_modeller_config.task.model.name
            if model_name:
                assert PyTorchTabularClassifier == eval(model_name)
                local_model = PyTorchTabularClassifier(
                    datastructure=datastructure,
                    schema=schema,
                    **hyperparameters,
                )

                local_results = local_model.fit(data=datasource)
                if local_results:
                    # Check results are the same
                    for k, pod_value in pod_results:
                        assert pod_value == local_results.get(k)
                else:
                    raise ValueError("Local results are None")
        finally:
            # Stop all processes
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                modeller_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            except Exception:
                pass

    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
    async def test_census_income_e2e_results_only(
        self,
        mock_bitfount_session: Mock,
        modeller_config: ModellerConfig,
        modeller_pswd: str,
        pod_1_config: PodConfig,
        pod_1_pswd: str,
        pod_2_config: PodConfig,
        pod_2_pswd: str,
        tmp_path: Path,
        token_dir: Path,
    ) -> None:
        """Tests a full end-to-end run.

        End to end run using Census income dataset with key-based id verification.
        """
        try:
            # Load configs and tie together
            tie_together_configs(modeller_config, pod_1_config, pod_2_config)
            # Mock BitfountSession
            # Override default return_value with side_effect
            # get_bitfount_session() one. side_effect takes precedence
            # over return_value.
            # If you ever change the order that the processes start in,
            # then be sure to change the order of these...
            mock_bitfount_session.side_effect = [
                # pod 1 bitfount session
                get_bitfount_session(pod_1_config.username, pod_1_pswd, token_dir),
                # pod 2 bitfount session
                get_bitfount_session(pod_2_config.username, pod_2_pswd, token_dir),
                # modeller bitfount session
                get_bitfount_session(
                    modeller_config.modeller.username, modeller_pswd, token_dir
                ),
            ]
            # Load pods
            pod_1 = setup_pod_from_config(pod_1_config)
            pod_2 = setup_pod_from_config(pod_2_config)

            # Start pods
            logger.info("Spinning up pods... ")
            logger.info("Spinning up pod 1... ")
            pod_1_process = multiprocessing.Process(
                target=pod_start, name="Pod_1_Runner", args=(pod_1,)
            )
            pod_1_process.start()

            logger.info("Spinning up pod 2... ")
            pod_2_process = multiprocessing.Process(
                target=pod_start, name="Pod_2_Runner", args=(pod_2,)
            )
            pod_2_process.start()
            time.sleep(POD_STARTUP_SLEEP)  # give the pods time to spin up

            assert pod_1_process.is_alive()
            assert pod_2_process.is_alive()

            logger.info("Pods loaded.")

            # Sort proactive access to each pod
            logger.info(f"Granting proactive access to pod {pod_1_config.pod_id}...")
            grant_proactive_access(
                modeller_username=modeller_config.modeller.username,
                pod_id=pod_1_config.pod_id,
                role="general_modeller",
                # This is a WebdriverBitfountSession as we mocked it out
                pod_session=cast(WebdriverBitfountSession, pod_1._session),
            )
            logger.info(f"Granting proactive access to pod {pod_2_config.pod_id}...")
            grant_proactive_access(
                modeller_username=modeller_config.modeller.username,
                pod_id=pod_2_config.pod_id,
                role="general_modeller",
                # This is a WebdriverBitfountSession as we mocked it out
                pod_session=cast(WebdriverBitfountSession, pod_2._session),
            )

            # Load modeller
            # We do this after the pods have started, as we need the pods to have
            # published their public keys to the Hub.
            modeller, pod_identifiers = setup_modeller_from_config(modeller_config)
            model_out = tmp_path / "model.out"

            modeller_process = multiprocessing.Process(
                target=run_modeller_process,
                name="Modeller_Run",
                args=(
                    modeller,
                    pod_identifiers,
                    model_out,
                ),
            )

            # Start modeller
            logger.info("Starting Modeller...")
            modeller_process.start()
            time.sleep(MODELLER_STARTUP_SLEEP)  # give modeller time to spin up
            assert not model_out.exists()  # check doesn't exist at this point

            # Join modeller, wait for output for MODELLER_RUN_TIMEOUT seconds
            modeller_process.join(MODELLER_RUN_TIMEOUT)

            # Check modeller is done, pods still going
            logger.info("Modeller should be finished by this point.")
            assert not modeller_process.is_alive()
            assert (
                modeller_process.exitcode is not None and modeller_process.exitcode <= 0
            )

            # Check output file exists and has nonzero size
            assert model_out.exists()
            assert os.path.getsize(model_out) > 0

            # Ensure pods are still running
            assert pod_1_process.is_alive()
            assert pod_2_process.is_alive()

            # Stop all pods
            for process in [pod_1_process, pod_2_process]:
                process.terminate()
                process.join()

            # Ensure shutdown
            assert pod_1_process.exitcode is not None and pod_1_process.exitcode <= 0
            assert pod_2_process.exitcode is not None and pod_2_process.exitcode <= 0
        finally:
            try:
                # The unbound variable issue is handled by the NameError catching below.
                # noinspection PyUnboundLocalVariable
                private_key_file: Path = cast(
                    Path, modeller_config.modeller.private_key_file
                )
                os.remove(private_key_file)
            except FileNotFoundError:
                pass
            except NameError:
                # Means the modeller_config variable didn't even get created;
                # another exception has occurred that has booted us out of the
                # `try` block before it even got to there. We pass on NameError
                # to avoid masking whatever that underlying exception was.
                pass

            # Stop all processes
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                modeller_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_1_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_2_process.terminate()
            except Exception:
                pass



"""This file enables the github repo to be packaged into a wheel using `setuptools`."""
import runpy

from setuptools import find_packages, setup

# Import version information into current namespace
file_globals = runpy.run_path("bitfount/__version__.py")

# Get security constraints
with open("requirements/constraints-security.txt") as f:
    security_constraints = f.read().splitlines()

# Get compatibility constraints
with open("requirements/constraints-compatibility.txt") as f:
    compat_constraints = f.read().splitlines()

# Get install requirements
with open("requirements/requirements.in") as f:
    install_reqs = f.read().splitlines()

# Add the security and compatibility constraints to the main requirements
# and get rid of unnecessary lines.
install_reqs.extend(security_constraints)
install_reqs.extend(compat_constraints)
install_reqs = [
    line.split("#")[0].strip()
    for line in install_reqs
    if not line.startswith(("#", "-c", "-r"))
]

# Get testing requirements
with open("requirements/requirements-tutorial.txt") as f:
    tutorial_reqs = f.read().splitlines()

# Get testing requirements
with open("requirements/requirements-test.txt") as f:
    tests_reqs = f.read().splitlines()

# Get dev requirements
with open("requirements/requirements-dev.txt") as f:
    dev_reqs = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    author=file_globals["__author__"],
    author_email=file_globals["__author_email__"],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Security :: Cryptography",
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description=file_globals["__description__"],
    entry_points={
        "console_scripts": [
            "bitfount = scripts.script_runner:main",
        ]
    },
    extras_require={
        "tests": tests_reqs,
        "dev": dev_reqs,
        "tutorials": tutorial_reqs,
    },
    install_requires=install_reqs,
    keywords=["federated learning", "privacy", "AI", "machine learning"],
    license="Apache License 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=file_globals["__title__"],
    packages=find_packages(),
    package_data={"bitfount": ["py.typed"]},
    project_urls={
        "Documentation": "https://docs.bitfount.com/",
        "Homepage": "https://bitfount.com",
        "Source Code": "https://github.com/bitfount/bitfount/",
        "Hub": "https://hub.bitfount.com",
    },
    python_requires=">=3.8,<3.10",
    url=file_globals["__url__"],
    version=file_globals["__version__"],
)



"""Tests base protocols module."""
from abc import ABC
import importlib
import inspect
import pkgutil
from unittest.mock import Mock, create_autospec

from pytest import fixture

from bitfount.federated.aggregators.aggregator import Aggregator
import bitfount.federated.algorithms.model_algorithms
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithmFactory,
)
import bitfount.federated.protocols
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
)
from bitfount.federated.protocols.results_only import ResultsOnly
from tests.utils.helper import unit_test


@unit_test
class TestBase:
    """Test protocol base module."""

    @fixture
    def mock_aggregator(self) -> Mock:
        """Returns mock aggregator."""
        mock_aggregator: Mock = create_autospec(Aggregator, instance=True)
        return mock_aggregator

    def test_protocol_initialises_pretrained_file_on_modeller_side(
        self, mock_aggregator: Mock, mock_modeller_mailbox: Mock
    ) -> None:
        """Test all protocols initialise pretrained file on modeller side."""
        pretrained_file_path = "mock/path"

        model_alg_modules = pkgutil.walk_packages(
            path=bitfount.federated.algorithms.model_algorithms.__path__,
            prefix=bitfount.federated.algorithms.model_algorithms.__name__ + ".",
        )

        for module_info in model_alg_modules:

            for _, cls in inspect.getmembers(
                importlib.import_module(module_info.name), inspect.isclass
            ):
                if (
                    issubclass(cls, _BaseModelAlgorithmFactory)
                    and ABC not in cls.__bases__
                ):

                    federated_algorithm = cls(
                        model=Mock(), pretrained_file=pretrained_file_path
                    )

                    for protocol_cls in [ResultsOnly, FederatedAveraging]:

                        protocol_factory = protocol_cls(
                            algorithm=federated_algorithm,
                            aggregator=mock_aggregator,
                            steps_between_parameter_updates=2,
                        )
                        protocol = protocol_factory.modeller(
                            mailbox=mock_modeller_mailbox
                        )

                        assert hasattr(protocol.algorithm, "pretrained_file")
                        assert (
                            protocol.algorithm.pretrained_file == pretrained_file_path
                        )



"""Backend-agnostic models that have no extra requirements.

:::info

Models defined here cannot be trained in a federated manner.

:::
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
import logging
import os

# TODO: [BIT-987] Review use of pickle
import pickle  # nosec
from typing import Any, List, Literal, Mapping, Optional, Tuple, Union, cast

from marshmallow import fields
import numpy as np
import pandas as pd
from pandas import DataFrame
from sklearn.linear_model import LinearRegression as sklearnLinearRegression
from sklearn.linear_model import LogisticRegression as sklearnLogisticRegression
from sklearn.neighbors import KNeighborsClassifier
import statsmodels.api as sm

from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.metrics import Metric
from bitfount.models.base_models import ClassifierMixIn, RegressorMixIn, _BaseModel
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


@delegates()
class LogisticRegressionClassifier(ClassifierMixIn, _BaseModel):
    """Wrapper around `sklearn.linear_model.LogisticRegression` model.

    For more details on the parameters, go to the scikit-learn documentation.

    Args:
        inverse_regularisation: Inverse regularisation parameter. Defaults to 0.0001.
        max_steps: Maximum number of steps to take. Defaults to 10000.
        model_type: Type of solver to use. Defaults to "lbfgs".
        penalty: Penalty to use. Defaults to "l2".
        early_stopping_tolerance: Tolerance for early stopping. Defaults to 1e-05.
        verbose: Verbosity level. Defaults to 0.

    Attributes:
        inverse_regularisation: Inverse regularisation parameter.
        max_steps: Maximum number of steps to take.
        model_type: Type of solver to use.
        penalty: Penalty to use.
        early_stopping_tolerance: Tolerance for early stopping.
        verbose: Verbosity level.
    """

    fields_dict = {
        "inverse_regularisation": fields.Float(allow_none=True),
        "max_steps": fields.Integer(allow_none=True),
        "model_type": fields.String(allow_none=True),
        "penalty": fields.String(allow_none=True),
        "early_stopping_tolerance": fields.Float(allow_none=True),
        "verbose": fields.Integer(allow_none=True),
    }

    def __init__(
        self,
        inverse_regularisation: Optional[float] = None,
        max_steps: Optional[int] = None,
        model_type: Optional[str] = None,
        penalty: Optional[str] = None,
        early_stopping_tolerance: Optional[float] = None,
        verbose: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.inverse_regularisation = (
            0.0001 if inverse_regularisation is None else inverse_regularisation
        )
        self.max_steps = 10000 if max_steps is None else max_steps
        self.model_type = "lbfgs" if model_type is None else model_type
        self.penalty = "l2" if penalty is None else penalty
        self.early_stopping_tolerance = (
            1e-05 if early_stopping_tolerance is None else early_stopping_tolerance
        )
        self.verbose = 0 if verbose is None else verbose

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model to file with provided `filename`.

        Args:
            filename: Path to file to save serialized model.
        """
        with open(filename, "wb") as f:
            pickle.dump(self._model, f, protocol=pickle.HIGHEST_PROTOCOL)

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model.

        :::danger

        This should not be used on a model file that has been received across a
        trust boundary due to underlying use of `pickle`.

        :::

        Args:
            filename: Path to file containing serialized model.
        """
        with open(filename, "rb") as f:
            # TODO: [BIT-987] Review use of pickle.load()
            self._model = pickle.load(f)  # nosec

    def evaluate(
        self, test_dl: Optional[BitfountDataLoader] = None, *args: Any, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform inference on test set and return predictions and targets.

        Args:
            test_dl: Optional `BitfountDataLoader` object containing test data. If this
                is not provided, the test set from the `BaseSource` used to train the
                model is used if present.

        Returns:
            A tuple of numpy arrays containing the predicted and actual values.

        Raises:
            ValueError: If there is no test data to evaluate the model on
        """
        logger.info("Evaluating logistic regression classifier.")

        if test_dl is None:
            if isinstance(self.test_dl, BitfountDataLoader):
                test_dl = self.test_dl
            else:
                raise ValueError("There is no test data to evaluate the model on.")
        test_df = test_dl.get_x_dataframe()
        test_preds: np.ndarray = cast(
            sklearnLogisticRegression, self._model
        ).predict_proba(test_df)
        test_target = test_dl.get_y_dataframe().to_numpy()

        return test_preds, test_target

    def fit(
        self,
        data: Optional[BaseSource] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Trains a model using the training set provided by `data`.

        :::info

        The validation set in `data` is not used when training this model.

        :::

        Args:
            data: `BaseSource` object containing training data.
        """
        logger.info("Fitting logistic regression classifier.")
        if data:
            data.load_data()
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, datasource=data
        )

        if not self.initialised:
            self.initialise_model(data)
        self.set_number_of_classes(table_schema)
        self._set_dataloaders()
        model = sklearnLogisticRegression(
            C=self.inverse_regularisation,
            random_state=self.seed,
            max_iter=self.max_steps,
            solver=self.model_type,
            verbose=self.verbose,
        )
        model.fit(
            self.train_dl.get_x_dataframe(),
            self.train_dl.get_y_dataframe()[self.databunch.target].to_numpy(),
        )
        self._model = model


@delegates()
class RegBoostRegressor(RegressorMixIn, _BaseModel):
    """Gradient Boosted Linear Regression Model.

    Implementation of "RegBoost: a gradient boosted multivariate regression algorithm"
    by Li et al. (2020). For more details, see the paper:
    https://www.emerald.com/insight/content/doi/10.1108/IJCS-10-2019-0029/full/html

    Args:
        learning_rate: Learning rate for gradient boosting. Defaults to 0.1.
        max_depth: Maximum depth of tree (number of nodes between root and leaf).
            A depth of 0 is equivalent to a single Linear Regression model. Defaults to
                10.
        min_data_points_per_node: Minimum number of data points required to split a
            node. Defaults to 5.
        stepwise_regression: Whether stepwise regression should go "forward" or
            "backward". Defaults to "forward".
        stepwise_regression_threshold: Threshold for stepwise regression. Defaults to
            0.15.

    Attributes:
        learning_rate: Learning rate for gradient boosting.
        max_depth: Maximum depth of tree (number of nodes between root and leaf).
        min_data_points_per_node: Minimum number of data points required to split a
            node.
        stepwise_regression: Whether stepwise regression should go "forward" or
            "backward".
        stepwise_regression_threshold: Threshold for stepwise regression.
    """

    fields_dict = {
        "learning_rate": fields.Float(allow_none=True),
        "max_depth": fields.Integer(allow_none=True),
        "min_data_points_per_node": fields.Integer(allow_none=True),
        "stepwise_regression": fields.String(allow_none=True),
        "stepwise_regression_threshold": fields.Float(allow_None=True),
    }

    def __init__(
        self,
        learning_rate: float = 0.1,
        max_depth: int = 10,
        min_data_points_per_node: int = 5,
        stepwise_regression: Literal["forward", "backward"] = "forward",
        stepwise_regression_threshold: float = 0.15,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.learning_rate = learning_rate
        self.max_depth = max_depth
        self.min_data_points_per_node = min_data_points_per_node
        self.stepwise_regression = stepwise_regression
        self.stepwise_regression_threshold = stepwise_regression_threshold

    class _ModelTreeSide(Enum):
        """Represents the side which a node represents in ModelTree."""

        POSITIVE = auto()
        NEGATIVE = auto()

    @dataclass
    class _ModelTree:
        """Core RegBoost model. Binary decision Tree of Linear Regressors.

        Args:
            data_indices: np array of data indices
            depth: number of parent nodes above this node until the root node
            side: -ve or +ve
        """

        data_indices: np.ndarray
        depth: int
        side: Optional[RegBoostRegressor._ModelTreeSide] = None

        def __post_init__(self) -> None:
            self.model = sklearnLinearRegression()
            self.classifier: Optional[
                KNeighborsClassifier
            ] = None  # only used at inference time
            self.features: List[str] = []
            self.negatives: Optional[RegBoostRegressor._ModelTree] = None
            self.positives: Optional[RegBoostRegressor._ModelTree] = None

        def __str__(self) -> str:
            """Prints model structure depth first vertically."""
            ret = (
                "┃\t" * self.depth
                + "┣"
                + repr(
                    f"Depth: {self.depth}, "
                    + f"Features: {len(self.features)}, "
                    + f"Data: {len(self.data_indices)}"
                )
                + "\n"
            )
            for child in [self.negatives, self.positives]:
                if child is not None:
                    ret += child.__str__()
            return ret

        def fit(self, *args: Any, **kwargs: Any) -> sklearnLinearRegression:
            """Call self.model fit method."""
            return self.model.fit(*args, **kwargs)

        def predict(self, *args: Any, **kwargs: Any) -> np.ndarray:
            """Call self.model predict method."""
            return cast(np.ndarray, self.model.predict(*args, **kwargs))

        @property
        def is_leaf(self) -> bool:
            """Returns boolean indicating whether this node is a leaf node."""
            return self.positives is None and self.negatives is None

        def add_child(
            self,
            side: RegBoostRegressor._ModelTreeSide,
            data_indices: np.ndarray,
        ) -> RegBoostRegressor._ModelTree:
            """Add child node (either `negatives` or `positives`) and return it.

            Args:
                side: -ve or +ve
                data_indices: np array of data indices

            Returns:
                RegBoostRegressor._ModelTree: the child node just created
            """
            model = RegBoostRegressor._ModelTree(data_indices, self.depth + 1, side)
            if side == RegBoostRegressor._ModelTreeSide.NEGATIVE:
                self.negatives = model
            elif side == RegBoostRegressor._ModelTreeSide.POSITIVE:
                self.positives = model

            return model

    def _fit_model_tree(
        self,
        model: _ModelTree,
        parent_preds: Optional[np.ndarray] = None,
        parent_features: Optional[List[str]] = None,
    ) -> None:
        """Fits `model` recursively.

        Builds `model` tree recursively depth-first and fits the Linear Regression
        model at each node.
        """
        data = self.train_dl.get_x_dataframe()

        if isinstance(data, DataFrame):
            X = data.loc[model.data_indices]
        else:
            X, _ = data
            X = X.loc[model.data_indices]
        y = (
            self.train_dl.get_y_dataframe()[self.datastructure.target]
            .to_numpy()[model.data_indices]
            .astype(np.float32)
        )

        # Update target to be the residual from parent * learning rate
        if parent_preds is not None:
            y -= self.learning_rate * parent_preds

        # Subset features to same as parent
        if parent_features is not None:
            X = X[parent_features]

        # Perform stepwise regression feature selection if we have more than one feature
        if len(X.columns) > 1:
            model.features = self._perform_stepwise_regression(X, y)
            X = X[model.features]
        else:
            model.features = parent_features or list(X.columns)

        # Fit linear regression model
        model.fit(X, y)

        # If we haven't reached the maximum depth, create positive and negative children
        if model.depth < self.max_depth:
            preds = model.predict(X)

            for side, indices in zip(
                RegBoostRegressor._ModelTreeSide,
                [preds <= y, preds > y],
            ):
                # Only create children if there are enough data points for both children
                if (
                    indices.sum() >= self.min_data_points_per_node
                    and (len(preds) - indices.sum()) >= self.min_data_points_per_node
                ):
                    model_ = model.add_child(side, X.index[indices])

                    # Add parent predictions to current predictions
                    new_preds = preds[indices]
                    if parent_preds is not None:
                        new_preds += parent_preds[indices]

                    # Fit child model tree
                    self._fit_model_tree(model_, new_preds, model.features)

    def _perform_stepwise_regression(self, X: pd.DataFrame, y: np.ndarray) -> List[str]:
        """Performs either forward or backward stepwise regression.

        Args:
            X (pd.DataFrame): dataframe of features
            y (np.ndarray): target array

        Raises:
            ValueError: if stepwise regression direction not supported

        Returns:
            List[str]: list of new features
        """
        if self.stepwise_regression == "forward":
            features = self._forward_stepwise_regression(
                X, y, self.stepwise_regression_threshold
            )
        elif self.stepwise_regression == "backward":
            features = self._backward_stepwise_regression(
                X, y, self.stepwise_regression_threshold
            )
        else:
            raise ValueError(
                "Stepwise regression only supports 'forward' and 'backward'"
            )
        return features

    @staticmethod
    def _forward_stepwise_regression(
        X: pd.DataFrame, y: np.ndarray, p_threshold: float
    ) -> List[str]:
        """Perform forward stepwise regression.

        Starts off with no features and keeps adding them until we reach p_threshold.
        """
        included: List[str] = []
        while True:
            if len(included) == len(X.columns):
                break

            p_values = {}
            for feature in [i for i in X.columns if i not in included]:
                model = sm.OLS(
                    y, sm.add_constant(pd.DataFrame(X[included + [feature]]))
                )
                results = model.fit()
                p_values[feature] = results.pvalues[feature]
            getter: Any = p_values.get
            best_feature = min(p_values, key=getter)
            best_pval = p_values[best_feature]

            if best_pval > p_threshold:
                break

            included.append(best_feature)

        return included or [best_feature]

    @staticmethod
    def _backward_stepwise_regression(
        X: pd.DataFrame, y: np.ndarray, p_threshold: float
    ) -> List[str]:
        """Perform backward stepwise regression.

        Starts off with all features and keeps removing them until we reach p_threshold.
        """
        included = list(X.columns)
        while True:
            if len(included) == 1:
                break

            model = sm.OLS(y, sm.add_constant(X[included]))
            results = model.fit()

            # use all coefs except intercept
            p_values = dict(results.pvalues.iloc[1:])
            getter: Any = p_values.get
            worst_feature = max(p_values, key=getter)
            worst_pval = p_values[worst_feature]

            if worst_pval <= p_threshold:
                break

            included.remove(worst_feature)

        return included

    def fit(
        self,
        data: Optional[BaseSource] = None,
        metrics: Optional[Mapping[str, Metric]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Trains a model using the training set provided by the BaseSource object."""
        logger.info("Fitting RegBoost regressor.")
        self.initialise_model(data)
        self._set_dataloaders()
        model = self._ModelTree(np.array(range(len(self.train_dl))), 0)
        self._fit_model_tree(model)
        self._model: RegBoostRegressor._ModelTree = model

    def evaluate(
        self, test_dl: Optional[BitfountDataLoader] = None, *args: Any, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform inference on test set and return predictions and targets.

        Args:
            test_dl: Optional `BitfountDataLoader` object containing test data. If this
                is not provided, the test set from the `BaseSource` used to train the
                model is used if present.

        Returns:
            A tuple of numpy arrays containing the predicted and actual values.

        Raises:
            ValueError: If there is no test data to evaluate the model on
        """
        logger.info("Evaluating RegBoost regressor.")

        k = cast(int, kwargs.get("k"))
        if test_dl is None:
            if isinstance(self.test_dl, BitfountDataLoader):
                test_dl = self.test_dl
            else:
                raise ValueError("There is no test data to evaluate the model on.")

        test_df = test_dl.get_x_dataframe()
        if isinstance(test_df, tuple):
            raise ValueError(
                "Multiple dataframes retrieved unexpectedly; "
                "this model does not support combination tabular and image data."
            )

        test_target = (
            test_dl.get_y_dataframe()[self.databunch.target]
            .to_numpy()
            .astype(np.float32)
        )
        test_preds = self._evaluate_model_tree(self._model, test_df, k)
        test_preds_aggregated = self._aggregate_model_predictions(test_preds)

        return test_preds_aggregated, test_target

    def _evaluate_model_tree(
        self, model: _ModelTree, test_df: pd.DataFrame, k: int
    ) -> List[List[float]]:
        """Recursively run inference on ModelTree and return final predictions."""
        if model.classifier is None:
            self._fit_model_tree_classifier(model, k)
        test_df = test_df[model.features]
        tree_classes = self._eval_model_tree_classifier(model, test_df)
        preds = np.reshape(model.predict(test_df), (len(test_df), 1)).tolist()

        # Split data into positives and negatives based on classifier results
        for indices, model_ in zip(
            [tree_classes == 1, tree_classes == 0],
            [model.positives, model.negatives],
        ):
            # If child model exists and there is at least one data point for that branch
            # Evaluate on that branch and append predictions to parent predictions
            if model_ is not None and sum(indices) > 0:
                child_preds = self._evaluate_model_tree(model_, test_df[indices], k)
                preds = self._append_model_predictions(preds, child_preds, indices)
        predictions: List[List[float]] = preds
        return predictions

    def _aggregate_model_predictions(self, preds: List[List[float]]) -> np.ndarray:
        """Aggregate all model predictions for each data point and return predictions.

        Learning rate applied to all predictions apart from the last one. This modifies
        the `preds` object.
        """
        preds_aggregated = []
        # pred_list: List[float]
        for pred_list in preds:
            preds_aggregated.append(
                pred_list[-1] + (sum(pred_list[:-1]) * self.learning_rate)
            )
        predictions: np.ndarray = np.asarray(preds_aggregated)
        return predictions

    @staticmethod
    def _append_model_predictions(
        parent_preds: List[List[float]],
        child_preds: List[List[float]],
        indices: List[bool],
    ) -> List[List[float]]:
        """Appends `child_preds` to corresponding `indices` in `parent_preds`."""
        i = 0
        for pred, idx in zip(parent_preds, indices):
            if idx:
                pred.extend(child_preds[i])
                i += 1
        return parent_preds

    def _fit_model_tree_classifier(self, model: _ModelTree, k: int) -> None:
        """Fit ModelTree classifier with training data."""
        if model.negatives is not None and model.positives is not None:
            model.classifier = KNeighborsClassifier(n_neighbors=k)

            X = self.train_dl.get_x_dataframe()
            if isinstance(X, tuple):
                raise ValueError(
                    "Multiple dataframes retrieved unexpectedly; "
                    "this model does not support combination tabular and image data."
                )

            X = X[model.features].loc[
                np.concatenate(
                    [model.positives.data_indices, model.negatives.data_indices]
                )
            ]
            y = [1 for _ in model.positives.data_indices] + [
                0 for _ in model.negatives.data_indices
            ]
            if isinstance(model.classifier, KNeighborsClassifier):
                model.classifier.fit(X, y)

    @staticmethod
    def _eval_model_tree_classifier(
        model: _ModelTree, test_df: pd.DataFrame
    ) -> np.ndarray:
        """Run inference on ModelTree classifier to return corresponding classes.

        Doesn't run the model if there is only one class available.
        """
        if isinstance(model.classifier, KNeighborsClassifier):
            class_predictions = model.classifier.predict(test_df)
        elif model.negatives is None:
            class_predictions = [1 for _ in range(len(test_df))]
        elif model.positives is None:
            class_predictions = [0 for _ in range(len(test_df))]

        return np.array(class_predictions)

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model to file with provided `filename`.

        Args:
            filename: Path to file to save serialized model.
        """
        with open(filename, "wb") as f:
            pickle.dump(self._model, f, protocol=pickle.HIGHEST_PROTOCOL)

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model.

        :::danger

        This should not be used on a model file that has been received across a
        trust boundary due to underlying use of `pickle`.

        :::

        Args:
            filename: Path to file containing serialized model.
        """
        self._set_dataloaders()
        with open(filename, "rb") as f:
            # TODO: [BIT-987] Review use of pickle.load()
            self._model = pickle.load(f)  # nosec



"""Utility functions for end-to-end testing."""
from collections import defaultdict
import inspect
import logging
from multiprocessing import current_process
import os
from pathlib import Path
from queue import Queue
import tempfile
import time
from typing import Any, Dict, Iterable, List, Optional, Tuple

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import desert
import yaml

from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.modeller import _Modeller
from bitfount.federated.pod import Pod
from bitfount.runners.config_schemas import ModellerConfig, PathConfig, PodConfig
from bitfount.runners.modeller_runner import DEFAULT_MODEL_OUT, run_modeller

logger = logging.getLogger(__name__)


def load_pod_config(config_path: Path, data_path: Path) -> PodConfig:
    """Loads a pod config and fixes up.

    Loads a pod config file, performing the necessary dynamic suffixing and
    envvar loading.
    """
    # Load base config
    with open(config_path) as f:
        config_yaml = yaml.safe_load(f)
    config: PodConfig = desert.schema(PodConfig).load(config_yaml)

    suffix = str(int(time.time()))  # unix timestamp

    # Add suffix where needed
    config.name += suffix
    if config.pod_details_config:
        config.pod_details_config.display_name += suffix
    if config.data_config.force_stypes:
        updated_force_stypes = {}
        for table_name in config.data_config.force_stypes:
            updated_force_stypes[
                f"{table_name}{suffix}"
            ] = config.data_config.force_stypes[table_name]
        config.data_config.force_stypes = updated_force_stypes
    # Update data
    config.data_config.datasource_args["path"] = PathConfig(data_path).path

    return config


def password_from_pod_config(config: PodConfig) -> str:
    """Get the user's hub password using the information in config."""
    return os.environ[f"{config.username}_pswd"]


def load_modeller_config(
    config_path: Path, private_key: Optional[RSAPrivateKey] = None
) -> ModellerConfig:
    """Loads modeller config and fixes up.

    Loads a modeller config file, performing the necessary dynamic suffixing,
    envvar loading, and private key substitution.
    """
    # Load base config
    with open(config_path) as f:
        config_yaml = yaml.safe_load(f)
    config: ModellerConfig = desert.schema(ModellerConfig).load(config_yaml)

    if not private_key:
        # Load appropriate private key from secrets onto the config
        private_key_str: str = os.environ[f"{config.modeller.username}_privkey"]
        private_key = _RSAEncryption.load_private_key(private_key_str.encode())
    config = _load_private_key_to_modeller_config(config, private_key)

    return config


def _load_private_key_to_modeller_config(
    config: ModellerConfig, private_key: RSAPrivateKey
) -> ModellerConfig:
    """Loads the private key as a file on the config.

    The caller is responsible for deleting the temporary file created.
    """
    # Create location for private key file. This is done separately to loading
    # the config because the key itself needs to be provided to write out to file
    # but we need the config to be loaded to extract the username.
    private_key_file = tempfile.NamedTemporaryFile(delete=False)
    private_key_file.write(_RSAEncryption.serialize_private_key(private_key))
    private_key_file.close()  # so can open it again later
    config.modeller.private_key_file = Path(private_key_file.name)
    return config


def password_from_modeller_config(config: ModellerConfig) -> str:
    """Get the modeller's hub password using the information in config."""
    return os.environ[f"{config.modeller.username}_pswd"]


def tie_together_configs(
    modeller_config: ModellerConfig, *pod_configs: PodConfig
) -> None:
    """Tie a set of modeller configs and pod configs with dynamic names together."""
    # Attach pods to modeller pod_names
    pod_ids = [p.pod_id for p in pod_configs]
    modeller_config.pods.identifiers = pod_ids

    # Attach pods to each others' "other_pods"
    for pod in pod_configs:
        other_pods = [p.pod_id for p in pod_configs if p is not pod]
        pod.approved_pods = other_pods


def get_caplog_records(queue: Queue) -> Dict[str, List[str]]:
    """Returns caplog records as dictionary of levels and messages."""
    records: Dict[str, List[str]] = defaultdict(list)
    while not queue.empty():
        record = queue.get()
        records[record.levelname].append(record.message)
    return records


def pod_start(pod: Pod) -> None:
    """Helper function for running pod in a separate process.

    Fails tests if there are any errors.
    """
    try:
        pod.start()
    except Exception as e:
        logger.error(f"Caught exception in {current_process().name} process: {e}")
        raise e


def run_modeller_process(
    modeller: _Modeller,
    pod_identifiers: Iterable[str],
    require_all_pods: bool = False,
    model_out: Path = DEFAULT_MODEL_OUT,
    queue: Optional[Queue] = None,
) -> Optional[Any]:
    """Helper function for sending Modeller requests and running training.

    Helper function for sending modeller training requests and running training
    in a separate process and failing test if there are any errors.

    Args:
        modeller (Modeller): The Modeller instance being used to manage the task.
        pod_identifiers (List[str]): List of pod identifiers to run the task against.
        require_all_pods: If true, raise PodResponseError if alteast one pod identifier
            specified rejects or fails to respond to a task request.
        model_out (Path): The path to save the model out to.
        queue (Queue): Optional queue to store output from running modeller process
    """
    try:
        res = run_modeller(modeller, pod_identifiers, require_all_pods, model_out)
        if res and queue:
            queue.put(res)
        return res
    except Exception as e:
        logger.error(f"Caught exception in {current_process().name} process: {e}")
        raise e


def _patch_authentication_code(
    username: str, password: str, token_dir: Path
) -> Tuple[Any, Any]:
    """Returns patchers for BitfountSession and OIDC authentication flow.

    Args:
        username (str): username
        password (str): password
        token_dir (Path): path to where `.token` file should be stored
    """
    oidc_webbrowser_import = (
        "bitfount.federated.transport.identity_verification"
        ".oidc.webbrowser.open_new_tab"
    )
    from functools import partial
    from pathlib import Path
    from unittest.mock import patch

    from tests.integration.bitfount_web_interactions import (
        get_bitfount_session,
        oidc_flow,
    )

    session_patcher = patch(
        "bitfount.hub.helper.BitfountSession",
        return_value=get_bitfount_session(
            username,
            password,
            Path(str(token_dir)),
        ),
    )
    oidc_patcher = patch(
        oidc_webbrowser_import,
        side_effect=partial(
            oidc_flow,
            username=username,
            password=password,
        ),
    )
    return session_patcher, oidc_patcher


def get_patched_authentication_code_nb(
    username: str, password: str, token_dir: Path
) -> str:
    """Returns code string for patching out normal authentication flow.

    Patches out BitfountSession as well as OIDC flow replacing it with a headless
    mechanism for doing OAuth verification.

    Args:
        username (str): username
        password (str): password
        token_dir (Path): path to where `.token` file should be stored
    """
    patch_auth_code = inspect.getsource(_patch_authentication_code)
    ext = (
        "session_patcher, oidc_patcher = _patch_authentication_code("
        f"'{username}', '{password}', "
        f"'{str(token_dir)}')\n"
        f"user_storage_path = Path('{str(token_dir)}')/'pods'\n"
        "session_patcher.start()\n"
        "oidc_patcher.start()\n"
    )
    patch_auth_code = patch_auth_code + ext
    return patch_auth_code


def rm_tree(pth: Path) -> None:
    """Removes all files from given folder and the folder."""
    for child in pth.iterdir():
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    pth.rmdir()



"""Contains models for running on the Bitfount platform."""
from typing import List

from bitfount.models.base_models import (
    ClassifierMixIn,
    CNNModelStructure,
    EarlyStopping,
    FeedForwardModelStructure,
    LoggerConfig,
    ModelContext,
    NeuralNetworkMixIn,
    NeuralNetworkModelStructure,
    NeuralNetworkPredefinedModel,
    Optimizer,
    RegressorMixIn,
    Scheduler,
)
from bitfount.models.bitfount_model import BitfountModel
from bitfount.models.models import LogisticRegressionClassifier, RegBoostRegressor

__all__: List[str] = [
    "BitfountModel",
    "ClassifierMixIn",
    "CNNModelStructure",
    "EarlyStopping",
    "FeedForwardModelStructure",
    "LoggerConfig",
    "LogisticRegressionClassifier",
    "ModelContext",
    "NeuralNetworkMixIn",
    "NeuralNetworkModelStructure",
    "NeuralNetworkPredefinedModel",
    "Optimizer",
    "RegBoostRegressor",
    "RegressorMixIn",
    "Scheduler",
]

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Test config settings for the end-to-end tests."""
import logging
from logging import handlers
from multiprocessing import Manager
from multiprocessing.managers import SyncManager
from pathlib import Path
import queue
from typing import Generator

from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempPathFactory
import dotenv
from pytest import fixture
import requests

import bitfount.config
from tests.integration import PRIVATE_DETAILS_ENV_FILE

logger = logging.getLogger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950

TMP_DIR_BASENAME = "E2E-"
CENSUS_INCOME_CSV_URL = (
    "https://bitfount-hosted-downloads.s3.eu-west-2.amazonaws.com/"
    "bitfount-tutorials/census_income.csv"
)


@fixture(autouse=True)
def env_fix(monkeypatch: MonkeyPatch) -> None:
    """Fix the environment into a known state for tests."""
    # Overrides the default fixture in tests/conftest.py
    monkeypatch.setenv("BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE)
    monkeypatch.setattr(
        "bitfount.config.BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE
    )


@fixture(scope="package", autouse=True)
def load_env() -> None:
    """Load the private environment variables file if it exists."""
    if PRIVATE_DETAILS_ENV_FILE.exists():
        dotenv.load_dotenv(dotenv_path=PRIVATE_DETAILS_ENV_FILE)


@fixture(scope="session")
def census_income_data(tmp_path_factory: TempPathFactory) -> Path:
    """Load the census income data from S3."""
    logging.info("Downloading census income data... ")
    tmp_dir = tmp_path_factory.mktemp(TMP_DIR_BASENAME)
    local_filename = tmp_dir / "census_income.csv"
    with requests.get(CENSUS_INCOME_CSV_URL) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            f.write(r.content)
    logging.info(f"Census income data saved to {local_filename}")
    return local_filename


@fixture(scope="module")
def manager() -> Generator[SyncManager, None, None]:
    """Creates a multiprocessing manager.

    Returns: Manager for use in tests
    """
    yield Manager()


@fixture
def caplog_queue(manager: SyncManager) -> Generator[queue.Queue, None, None]:
    """Capture logs from other processes in a Queue and returns it."""
    logger_queue = manager.Queue()
    # We want to capture all logs, not just ours, so use root logger
    logger_for_caplog = logging.getLogger()
    logger_for_caplog.setLevel(logging.INFO)
    logger_for_caplog.addHandler(handlers.QueueHandler(logger_queue))
    yield logger_queue



"""Base classes for all model-based algorithms.

Attributes:
    registry: A read-only dictionary of model algorithm factory names to their
        implementation classes.
"""
from __future__ import annotations

from abc import ABC
import inspect
import os
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    Mapping,
    Optional,
    Type,
    Union,
    cast,
)

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseAlgorithmFactory,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.models.base_models import MAIN_MODEL_REGISTRY, ModelContext
from bitfount.schemas.utils import bf_dump
from bitfount.types import (
    T_DTYPE,
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    DistributedModelProtocol,
    _SerializedWeights,
    _StrAnyDict,
)

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub
    from bitfount.types import _DistributedModelTypeOrReference


logger = _get_federated_logger(__name__)


class _BaseModelAlgorithm(Generic[T_DTYPE], _BaseAlgorithm, ABC):
    """Blueprint for either the modeller side or the worker side of ModelAlgorithm."""

    def __init__(
        self,
        *,
        model: DistributedModelProtocol,
        pretrained_file: Optional[Union[str, os.PathLike]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.pretrained_file = pretrained_file

    @property
    def tensor_precision(self) -> T_DTYPE:
        """Returns model tensor precision."""
        return self.model.tensor_precision()


class _BaseModellerModelAlgorithm(_BaseModelAlgorithm, _BaseModellerAlgorithm, ABC):
    """Modeller side of the algorithm."""

    def __init__(
        self,
        *,
        model: DistributedModelProtocol,
        pretrained_file: Optional[Union[str, os.PathLike]] = None,
        **kwargs: Any,
    ):
        super().__init__(model=model, pretrained_file=pretrained_file, **kwargs)

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Initialises the algorithm as required."""
        if not self.model.initialised:
            self.model.initialise_model(context=ModelContext.MODELLER)

        # This needs to occur AFTER model initialization so the model is correctly
        # created. deserialize() may cause initialization but we can not rely on it
        # in this instance because we need to pass in context information.
        # This should be reviewed as part of [BIT-536].
        if self.pretrained_file is not None:
            logger.info(f"Deserializing model from {self.pretrained_file}.")
            self.model.deserialize(self.pretrained_file)


class _BaseWorkerModelAlgorithm(_BaseModelAlgorithm, _BaseWorkerAlgorithm, ABC):
    """Worker side of the algorithm."""

    def __init__(self, *, model: DistributedModelProtocol, **kwargs: Any):
        super().__init__(model=model, **kwargs)

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        model_params: Optional[_SerializedWeights] = None,
        **kwargs: Any,
    ) -> None:
        """Initialises the algorithm as required."""
        # Apply pod DP settings if needed. Needs to occur before model
        # initialization so the right DP settings are applied during initialization.

        self._apply_pod_dp(pod_dp)
        self.model.initialise_model(data=datasource, context=ModelContext.WORKER)
        if model_params:
            tensor_model_params = self.model.deserialize_params(model_params)
            self.model.update_params(tensor_model_params)

    def _apply_pod_dp(self, pod_dp: Optional[DPPodConfig]) -> None:
        """Applies pod-level Differential Privacy constraints if supported.

        The model must inherit from `DifferentiallyPrivate` for DP to be supported.

        Args:
            pod_dp: The pod DP constraints to apply or None if no constraints.
        """
        try:
            # only applied if model supports DP so can ignore attr-defined
            self.model.apply_pod_dp(pod_dp)  # type: ignore[attr-defined]  # Reason: caught by try-except  # noqa: B950
        except AttributeError:
            pass


# The mutable underlying dict that holds the registry information
_registry: Dict[str, Type[_BaseModelAlgorithmFactory]] = {}
# The read-only version of the registry that is allowed to be imported
registry: Mapping[str, Type[_BaseModelAlgorithmFactory]] = MappingProxyType(_registry)


class _BaseModelAlgorithmFactory(_BaseAlgorithmFactory, ABC):
    """Base factory for algorithms involving an underlying model.

    Args:
        model: The model for the federated algorithm.
        pretrained_file: A file path or a string containing a
            pre-trained model. Defaults to None.

    Attributes:
        model: The model for the federated algorithm.
        pretrained_file: A file path or a string containing a
            pre-trained model. Defaults to None.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"model": MAIN_MODEL_REGISTRY}

    def __init__(
        self,
        *,
        model: _DistributedModelTypeOrReference,
        pretrained_file: Optional[Union[str, os.PathLike]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.pretrained_file = pretrained_file

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to Model Algorithm registry")
            _registry[cls.__name__] = cls

    @property
    def model_schema(self) -> _StrAnyDict:
        """Returns underlying model Schema."""
        self.model = cast(DistributedModelProtocol, self.model)
        return bf_dump(self.model)

    def _get_model_from_reference(
        self, hub: Optional[BitfountHub] = None
    ) -> DistributedModelProtocol:
        """Returns underlying model if BitfountModelReference.

        If not, just returns self.model.
        """
        # TODO: [BIT-890] perhaps move this logic one level higher so that the algorithm
        # factory always takes a DistributedModelProtocol
        if isinstance(self.model, BitfountModelReference):
            if hub is not None:
                self.model.hub = hub
            model = self.model.get_model()(
                datastructure=self.model.datastructure,
                schema=self.model.schema,
                **self.model.hyperparameters,
            )
            self.model = cast(DistributedModelProtocol, model)
            return self.model
        else:
            return self.model



"""Secure model parameter aggregators for Federated Averaging."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, List, Mapping, Optional, Union, cast, overload

import numpy as np
import pandas as pd

from bitfount.federated.aggregators.aggregator import _ModellerSide as ModellerSide_
from bitfount.federated.aggregators.aggregator import _WorkerSide as WorkerSide_
from bitfount.federated.aggregators.base import AggregationType, _BaseAggregatorFactory
from bitfount.federated.exceptions import AggregatorError
from bitfount.federated.secure import SecureShare, _secure_share_registry
from bitfount.federated.shim import BackendTensorShim, _load_default_tensor_shim
from bitfount.federated.task_requests import _TaskRequest
from bitfount.federated.transport.worker_transport import _InterPodWorkerMailbox
from bitfount.federated.types import AggregatorType
from bitfount.types import (
    T_DTYPE,
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _SerializedWeights,
    _Weights,
)
from bitfount.utils import delegates


class _BaseSecureAggregator:
    """Shared behaviour and attributes for SecureAggregator classes."""

    def __init__(self, *, secure_share: SecureShare, **kwargs: Any):
        self.secure_share = secure_share
        super().__init__(**kwargs)


class _ModellerSide(_BaseSecureAggregator, ModellerSide_):
    """Modeller-side of the secure aggregator."""

    def __init__(
        self,
        *,
        secure_share: SecureShare,
        tensor_shim: BackendTensorShim,
        **kwargs: Any,
    ):
        # SecureAggregation not yet compatible with update weighting; need to
        # check it hasn't been supplied.
        # TODO: [BIT-1486] Remove this constraint
        try:
            if kwargs["weights"] is not None:
                raise NotImplementedError(
                    "SecureAggregation does not support update weighting"
                )
        except KeyError:
            pass

        super().__init__(secure_share=secure_share, tensor_shim=tensor_shim, **kwargs)

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, _SerializedWeights],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> _Weights:
        """Decodes and averages tensor dictionaries and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, pd.DataFrame],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Decodes and averages pandas dataframes and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, np.ndarray],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """Decodes and averages numpy arrays and returns the average."""
        ...

    def run(
        self,
        algorithm_outputs: Union[
            Mapping[str, _SerializedWeights],
            Mapping[str, np.ndarray],
            Mapping[str, pd.DataFrame],
        ],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> Union[_Weights, np.ndarray, pd.DataFrame]:
        """Decodes and averages algorithm outputs and returns them.

        Args:
            algorithm_outputs: A mapping of pod names to their algorithm outputs.
            tensor_dtype: The dtype to use when converting parameter updates to tensors.

        Raises:
            AggregatorError: If there is any issue whereby aggregation cannot proceed.
                Details will be included in the error message.
        """
        obj_type = self._get_type(algorithm_outputs.values())
        self._validate_algorithm_outputs(algorithm_outputs.values(), obj_type)
        averaged_output: Union[pd.DataFrame, np.ndarray, _Weights]

        if obj_type == AggregationType.TENSOR_DICT:
            # First we convert the parameter updates to numpy to allow them to be
            # decoded more easily.
            algorithm_outputs = cast(
                Mapping[str, _SerializedWeights], algorithm_outputs
            )
            new_parameter_updates: List[Dict[str, np.ndarray]] = []
            for update in algorithm_outputs.values():
                new_update: Dict[str, np.ndarray] = {}
                for name, param in update.items():
                    new_update[name] = self._tensor_shim.to_numpy(param)
                new_parameter_updates.append(new_update)

            # Then we average and decode the parameters.
            average_state_dict = self.secure_share.average_and_decode_state_dicts(
                new_parameter_updates
            )

            # Finally we convert the averaged parameters back to tensors.
            averaged_output = {}
            for param_name, param_ in average_state_dict.items():
                averaged_output[param_name] = self._tensor_shim.to_tensor(
                    param_, dtype=tensor_dtype
                ).squeeze()

        elif obj_type == AggregationType.PANDAS_DATAFRAME:
            # First we convert the dataframes to dictionaries of numpy arrays to allow
            # them to be decoded more easily.
            algorithm_outputs = cast(Mapping[str, pd.DataFrame], algorithm_outputs)
            algorithm_output_values = [
                {
                    col: output[col].to_numpy()
                    for col in output.columns
                    if col not in self.non_numeric_columns
                }
                for output in algorithm_outputs.values()
            ]
            # Average and decode the numeric columns in the dataframes
            averaged_columns: Dict[
                str, np.ndarray
            ] = self.secure_share.average_and_decode_state_dicts(
                algorithm_output_values
            )
            # Add back the non-numeric columns to the averaged columns.
            averaged_output = pd.DataFrame(
                {**self.non_numeric_columns, **averaged_columns}
            )
        elif obj_type == AggregationType.NUMPY_ARRAY:
            algorithm_outputs = cast(Mapping[str, np.ndarray], algorithm_outputs)
            averaged_numpy_array = self.secure_share.average_and_decode_state_dicts(
                [{"array": i} for i in algorithm_outputs.values()]
            )
            averaged_output = averaged_numpy_array["array"]

        return averaged_output


class _WorkerSide(_BaseSecureAggregator, WorkerSide_):
    """Worker-side of the secure aggregator."""

    def __init__(
        self,
        *,
        secure_share: SecureShare,
        mailbox: _InterPodWorkerMailbox,
        tensor_shim: BackendTensorShim,
        **kwargs: Any,
    ):
        super().__init__(secure_share=secure_share, tensor_shim=tensor_shim, **kwargs)
        self.mailbox = mailbox

    @overload
    async def run(
        self, algorithm_output: _Weights, **kwargs: Any
    ) -> _SerializedWeights:
        """Performs secure aggregation on the tensor dictionary.

        The tensors in the dictionary are then converted into a list of floats before
        being returned.
        """
        ...

    @overload
    async def run(self, algorithm_output: np.ndarray, **kwargs: Any) -> np.ndarray:
        """Performs secure aggregation on the array and returns it."""
        ...

    @overload
    async def run(self, algorithm_output: pd.DataFrame, **kwargs: Any) -> pd.DataFrame:
        """Performs secure aggregation on the dataframe and returns it."""
        ...

    async def run(
        self, algorithm_output: Union[_Weights, np.ndarray, pd.DataFrame], **kwargs: Any
    ) -> Union[_SerializedWeights, np.ndarray, pd.DataFrame]:
        """Encodes input, converts it to a serializable object, and returns it.

        Only tensor dictionaries are converted from tensors to lists of floats. Pandas
        dataframes and numpy arrays are left as they are.

        Args:
            algorithm_output: The algorithm output to be encoded and sent to the
                modeller.

        Raises:
            AggregatorError: If the algorithm output is not a supported type.
        """
        obj_type = self._get_type([algorithm_output])

        if obj_type == AggregationType.PANDAS_DATAFRAME:
            algorithm_output = cast(pd.DataFrame, algorithm_output)
            self._set_non_numeric_columns([algorithm_output])
            algorithm_output_np: Dict[str, np.ndarray] = {
                col: algorithm_output[col].to_numpy()
                for col in algorithm_output.columns
            }
            keep_columns = [
                i for i in algorithm_output_np if i not in self.non_numeric_columns
            ]
            secure_aggregation_output = await self.secure_share.do_secure_aggregation(
                {k: algorithm_output_np[k] for k in keep_columns},
                self.mailbox,
            )
            return pd.DataFrame(
                {**self.non_numeric_columns, **secure_aggregation_output}
            )

        elif obj_type == AggregationType.NUMPY_ARRAY:
            algorithm_output = cast(np.ndarray, algorithm_output)
            secure_aggregation_output = await self.secure_share.do_secure_aggregation(
                {"array": algorithm_output}, self.mailbox
            )
            return secure_aggregation_output["array"]
        elif obj_type == AggregationType.TENSOR_DICT:
            algorithm_output = cast(_Weights, algorithm_output)
            secure_aggregation_output = await self.secure_share.do_secure_aggregation(
                algorithm_output, self.mailbox
            )
            # We are reusing secure_algorithm_output and changing it to
            # SerializedWeights which is why we ignore the assignment issue.
            for name, param in secure_aggregation_output.items():
                secure_aggregation_output[name] = self._tensor_shim.to_list(param)  # type: ignore[assignment] # Reason: see comment # noqa: B950

            return cast(_SerializedWeights, secure_aggregation_output)

        raise AggregatorError("Unsupported object type for Secure Aggregation.")


class _InterPodAggregatorWorkerFactory(ABC):
    """Defines the worker() factory method for inter-pod aggregation."""

    @abstractmethod
    def worker(self, mailbox: _InterPodWorkerMailbox, **kwargs: Any) -> _WorkerSide:
        """Returns worker side of the SecureAggregator."""
        pass


@delegates()
class SecureAggregator(_BaseAggregatorFactory, _InterPodAggregatorWorkerFactory):
    """Secure aggregation of encrypted model parameters.

    This aggregator is used to aggregate model parameter updates from multiple workers
    without revealing the individual updates to any worker or the modeller. This is
    known as Secure Multi-Party Computation (SMPC).

    Args:
        secure_share: The `SecureShare` object used to encrypt and decrypt model
            parameters and communicate with the other workers.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"_secure_share": _secure_share_registry}

    def __init__(
        self,
        secure_share: Optional[SecureShare] = None,
        **kwargs: Any,
    ):

        super().__init__(**kwargs)
        self._tensor_shim = _load_default_tensor_shim()
        self._secure_share = secure_share if secure_share else SecureShare()

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the SecureAggregator."""
        return _ModellerSide(
            tensor_shim=self._tensor_shim, secure_share=self._secure_share, **kwargs
        )

    def worker(self, mailbox: _InterPodWorkerMailbox, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the SecureAggregator.

        Args:
            mailbox: The mailbox to use for receiving messages.
        """
        return _WorkerSide(
            tensor_shim=self._tensor_shim,
            secure_share=self._secure_share,
            mailbox=mailbox,
            **kwargs,
        )


def _is_secure_share_task_request(task_request: _TaskRequest) -> bool:
    """Checks if a task request is for secure share aggregation."""
    aggregator = task_request.serialized_protocol.get("aggregator")
    if aggregator:
        return aggregator["class_name"] == AggregatorType.SecureAggregator.value
    return False



"""Modeller for dispatching tasks."""
from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Type,
    Union,
    cast,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithmFactory,
)
from bitfount.federated.authorisation_checkers import (
    _IDENTITY_VERIFICATION_METHODS_MAP,
    IdentityVerificationMethod,
    _SignatureBasedAuthorisation,
    check_identity_verification_method,
)
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.exceptions import PodResponseError
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_message_service,
    _get_idp_url,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.monitoring import task_monitor_context
from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from bitfount.federated.transport.identity_verification.oidc import (
    _OIDCAuthFlowChallengeHandler,
    _OIDCDeviceCodeHandler,
)
from bitfount.federated.transport.identity_verification.saml import (
    _SAMLChallengeHandler,
)
from bitfount.federated.transport.identity_verification.types import (
    _HasWebServer,
    _ResponseHandler,
)
from bitfount.federated.transport.message_service import (
    _BitfountMessageType,
    _MessageService,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.types import _TaskRequestMessageGenerator
from bitfount.hub.api import BitfountHub
from bitfount.hub.authentication_flow import _get_auth_environment
from bitfount.hub.exceptions import PodDoesNotExistError
from bitfount.hub.helper import _default_bitfounthub, _get_pod_public_keys
from bitfount.types import DistributedModelProtocol

if TYPE_CHECKING:
    from bitfount.federated.protocols.base import _BaseProtocolFactory


logger = _get_federated_logger(__name__)


class _Modeller:
    """Dispatches tasks to pods and runs the modeller side of the provided protocol.

    ```python title="Example usage:"
    import bitfount as bf

    modeller=bf.Modeller(
        protocol=bf.FederatedAveraging(...),
    )
    modeller.run(pod_identifiers=["bitfount/example-pod-1", "bitfount/example-pod-2"])
    ```

    Args:
        protocol: The protocol to use for the task.
        message_service: The message service to use for communication with pods.
            Defaults to None, in which case a new message service will be created.
        bitfounthub: Hub instance for Bitfount Hub communication. Defaults to None,
            in which case a new `BitfountHub` instance will be created.
        pod_public_key_paths: Optional. Mapping of pod identifiers to public
            key files for existing pod public keys. Expired or non-existent
            keys will be downloaded from the Hub.
        identity_verification_method: The identity verification method to use.
            Defaults to OIDC_DEVICE_CODE.
        private_key: This modeller's private key either as an `RSAPrivateKey`
            instance or a path to a private key file. If a non-key-based identity
            verification method is used, this is ignored. Defaults to None.
        idp_url: URL of identity provider, used for the SAML Identity
            Verification Method.

    Attributes:
        protocol: The protocol to use for the task.

    Raises:
        ValueError: If key-based identity verification is selected but no
            private key is provided.
    """

    def __init__(
        self,
        protocol: _BaseProtocolFactory,
        message_service: Optional[_MessageService] = None,
        bitfounthub: Optional[BitfountHub] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        identity_verification_method: Union[
            str, IdentityVerificationMethod
        ] = IdentityVerificationMethod.DEFAULT,
        private_key: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
    ):
        self.protocol = protocol
        self._hub = _default_bitfounthub(hub=bitfounthub)
        self._message_service = (
            message_service
            if message_service is not None
            else _create_message_service(session=self._hub.session)
        )
        self._pod_public_key_paths: Mapping[str, Path] = pod_public_key_paths or dict()

        self._identity_verification_method: IdentityVerificationMethod = (
            check_identity_verification_method(identity_verification_method)
        )

        # Load private key if needed
        if isinstance(private_key, Path):
            private_key = _RSAEncryption.load_private_key(private_key)
        self._private_key = private_key

        self._idp_url: Optional[str]
        if self._identity_verification_method == IdentityVerificationMethod.SAML:
            # We also need to set the idp_url based on the environment if using SAML
            self._idp_url = idp_url if idp_url is not None else _get_idp_url()
        else:
            self._idp_url = None

    @property
    def _private_key(self) -> Optional[RSAPrivateKey]:
        """Private key attribute."""
        return self.__private_key

    @_private_key.setter
    def _private_key(self, private_key: Optional[RSAPrivateKey]) -> None:
        """Set _private_key attribute."""
        self.__private_key: Optional[RSAPrivateKey] = None

        if self._identity_verification_method == IdentityVerificationMethod.KEYS:
            if not private_key:
                raise ValueError(
                    "Key-based identity verification selected but "
                    "no private key provided."
                )
            self.__private_key = private_key
        else:
            if private_key:
                logger.warning(
                    f"Private key provided but identity verification method "
                    f'"{self._identity_verification_method.value}" was chosen. '
                    f"Private key will be ignored."
                )

    async def _send_task_requests(
        self, pod_identifiers: Iterable[str]
    ) -> _ModellerMailbox:
        """Sends task requests to pods.

        Args:
            pod_identifiers: The group of pods to run the task with.

        Returns:
            The created ModellerMailbox for the task.
        """
        # Load pod public keys
        loaded_pod_public_keys: Dict[str, RSAPublicKey] = _get_pod_public_keys(
            pod_identifiers=pod_identifiers,
            hub=self._hub,
            pod_public_key_paths=self._pod_public_key_paths,
        )

        task_request_msg_gen = self._get_task_request_msg_gen()
        # Send task requests to chosen pods. We haven't attached a handler for the
        # responses yet but this is handled below. There is no issue due to the
        # fact that the listening task hasn't been started yet, and the inherent
        # handler backoff.
        modeller_mailbox: _ModellerMailbox = await _ModellerMailbox.send_task_requests(
            serialized_protocol=self.protocol.dump(),
            pod_public_keys=loaded_pod_public_keys,
            task_request_msg_gen=task_request_msg_gen,
            message_service=self._message_service,
        )

        return modeller_mailbox

    def _get_task_request_msg_gen(self) -> _TaskRequestMessageGenerator:
        """Construct correct TaskRequestMessageGenerator object for the auth type."""
        authorization_checker_cls = _IDENTITY_VERIFICATION_METHODS_MAP[
            self._identity_verification_method
        ]

        if self._identity_verification_method == IdentityVerificationMethod.KEYS:
            authorization_checker_cls = cast(
                Type[_SignatureBasedAuthorisation], authorization_checker_cls
            )
            assert isinstance(self._private_key, RSAPrivateKey)  # nosec[assert_used]
            return authorization_checker_cls.create_task_request_message_generator(
                self._private_key
            )
        else:
            return authorization_checker_cls.create_task_request_message_generator()

    async def _modeller_run(
        self,
        modeller_mailbox: _ModellerMailbox,
        pod_identifiers: Iterable[str],
        require_all_pods: bool = False,
        response_handler: Optional[_ResponseHandler] = None,
    ) -> Union[Literal[False], Optional[Any]]:
        """Waits for pod responses and handles any who don't respond in time.

        Runs the modeller side of the protocol if some pods accepted the task.
        """
        # If additional response handling is specified, we call the handler
        if response_handler:
            await response_handler.handle(modeller_mailbox)

        # Process job accept/reject
        await modeller_mailbox.process_task_request_responses()

        # Fail-fast if no pods accepted the task request.
        if not modeller_mailbox.accepted_worker_mailboxes:
            logger.error(
                "No workers with which to train. "
                "Please ensure you have sent task requests at least some of "
                "which have been accepted."
            )
            return False

        # Otherwise run the protocol
        unaccepted_worker_ids = set(pod_identifiers).difference(
            set(modeller_mailbox.accepted_worker_mailboxes)
        )
        if unaccepted_worker_ids:
            unaccepted_worker_msg = (
                f"Pods {', '.join(unaccepted_worker_ids)} "
                "rejected task request or failed to respond. "
            )
            if require_all_pods:
                raise PodResponseError(
                    unaccepted_worker_msg
                    + "Task requires all pods accept the task request."
                )
            else:
                logger.warning(
                    unaccepted_worker_msg + "Continuing task without these pods ..."
                )

        await modeller_mailbox.send_task_start_message()

        return await self._run_modeller_protocol(modeller_mailbox)

    async def _run_modeller_protocol(
        self,
        modeller_mailbox: _ModellerMailbox,
    ) -> Optional[Any]:
        """Runs the modeller-side of the protocol."""
        protocol = self.protocol.modeller(mailbox=modeller_mailbox)
        # Run the task
        result = await protocol.run()
        return result

    async def run_async(
        self,
        pod_identifiers: Iterable[str],
        require_all_pods: bool,
    ) -> Union[Literal[False], Optional[Any]]:
        """Runs the modeller's task with a set of pods.

        Will send a task request before commencing the task itself and only pods that
        accept the task request will be used.

        Args:
            pod_identifiers: The identifiers for the pods to run the task on.

        Returns:
            Whatever the protocol's return value is.

        Raises:
            PodResponseError: If require_all_pods is True and at least one
                pod_identifier rejects or fails to respond to the task request.
        """
        try:
            # We do this early to give any handler components time to spin up
            response_handler = self._get_response_handler()

            logger.info("Sending task requests...")
            modeller_mailbox: _ModellerMailbox = await self._send_task_requests(
                pod_identifiers
            )

            with task_monitor_context(
                hub=self._hub,
                task_id=modeller_mailbox.task_id,
                sender_id=modeller_mailbox.mailbox_id,
            ):
                run_result: Union[
                    Literal[False], Optional[Any]
                ] = await _run_func_and_listen_to_mailbox(
                    self._modeller_run(
                        modeller_mailbox=modeller_mailbox,
                        pod_identifiers=pod_identifiers,
                        require_all_pods=require_all_pods,
                        response_handler=response_handler,
                    ),
                    modeller_mailbox,
                )
                modeller_mailbox.delete_all_handlers(_BitfountMessageType.LOG_MESSAGE)

                logger.info("Task complete")
                return run_result
        except PodDoesNotExistError as e:
            logger.exception(e)
            logger.error("Aborted task request.")
            return False
        finally:
            # Stop any response handler web servers that may be running still
            try:
                # noinspection PyUnboundLocalVariable
                await cast(_HasWebServer, response_handler).stop_server()
            except NameError:
                logger.warning("Tried to shutdown non-existent response handler")
            except AttributeError:
                # Didn't have a stop_server() method
                pass

    def _get_response_handler(self) -> Optional[_ResponseHandler]:
        """Construct a response handler for additional task messages if required.

        This may spin-up web servers if the response handler requires it. It is
        the responsibility of the handler to shut these down when no longer needed.
        """
        response_handler: Optional[_ResponseHandler] = None

        # SAML
        if self._identity_verification_method == IdentityVerificationMethod.SAML:
            logger.debug("Setting up SAML challenge listener")
            # If SAML is used, the url is set in the modeller init
            self._idp_url = cast(str, self._idp_url)
            response_handler = _SAMLChallengeHandler(self._idp_url)

        # OIDC_ACF_PKCE
        elif (
            self._identity_verification_method
            == IdentityVerificationMethod.OIDC_ACF_PKCE
        ):
            auth_env = _get_auth_environment()
            logger.debug(
                f"Setting up OIDC Authorization Code Flow challenge listener against "
                f"{auth_env.name} authorization environment."
            )
            response_handler = _OIDCAuthFlowChallengeHandler(
                auth_domain=auth_env.auth_domain
            )

        # OIDC_DEVICE_CODE
        elif (
            self._identity_verification_method
            == IdentityVerificationMethod.OIDC_DEVICE_CODE
        ):
            auth_env = _get_auth_environment()
            response_handler = _OIDCDeviceCodeHandler(auth_domain=auth_env.auth_domain)

        # Start web server if needed
        # cast()s are needed to allow mypy to infer this as reachable
        if isinstance(cast(_HasWebServer, response_handler), _HasWebServer):
            # Start response handler web server in background as it takes some time
            cast(_HasWebServer, response_handler).start_server()

        return response_handler

    def _serialize(self, model_out: Union[str, os.PathLike]) -> None:
        """Save model to file.

        Save the model to the file specified.

        Args:
            model_out: The path to save the model out to.
        """
        if self.protocol.class_name != "ResultsOnly":
            # If protocol is not ResultsOnly, then the algorithm has a model
            # Assertion to reassure mypy of this fact
            assert isinstance(  # nosec[assert_used]
                self.protocol.algorithm, _BaseModelAlgorithmFactory
            )
            self.protocol.algorithm.model = cast(
                DistributedModelProtocol, self.protocol.algorithm.model
            )
            logger.info(f'Saving model out to "{model_out}"')
            self.protocol.algorithm.model.serialize(model_out)
        else:
            logger.warning(
                "Model not saved. ResultsOnly protocol does not produce model."
            )

    def run(
        self,
        pod_identifiers: Iterable[str],
        require_all_pods: bool = False,
        model_out: Optional[Union[Path, str]] = None,
    ) -> Union[Literal[False], Optional[Any]]:
        """Runs the modeller's task with a set of pods.

        Will send a task request before commencing the task itself and only pods
        that accept the task request will be used.

        Args:
            pod_identifiers: The identifiers for the pods to run the task on.
            model_out: If specified, path to save the model out to.

        Returns:
            Whatever the protocol's return value is.

        Raises:
            PodResponseError: If require_all_pods is True and at least one
                pod_identifier rejects or fails to respond to the task request.
        """
        pod_identifiers = _check_and_update_pod_ids(pod_identifiers, self._hub)

        # We need to pass the `SecureShare` parameters to the
        # model when SecureAggregation is in use.
        # First we check the protocol's model and aggregators.
        # if steps_between_parameter_updates is 1, then we do
        # the clipping at the 'SecureShare' level.
        if (
            (
                hasattr(self.protocol, "aggregator")
                and isinstance(self.protocol.aggregator, SecureAggregator)  # type: ignore[attr-defined] # Reason: see below # noqa: B950
            )
            and (
                hasattr(self.protocol, "steps_between_parameter_updates")
                and self.protocol.steps_between_parameter_updates != 1  # type: ignore[attr-defined] # Reason: see below # noqa: B950
            )
            and hasattr(self.protocol.algorithm, "model")
        ):
            logger.warning(
                "SecureAggregation in use. "
                "We recommend normalization of continuous features prior to training."
            )
            # Show warning to user if using custom models that
            # they might have to implement clipping in their custom model.

            if isinstance(self.protocol.algorithm.model, BitfountModelReference):  # type: ignore[attr-defined] # Reason: see below # noqa: B950
                logger.warning(
                    "You are using a custom model with Secure Aggregation."
                    "We recommend clipping the model parameters."
                )
            # Mypy: hasattr check on the protocol aggregator, model and steps.
            # Pass the `SecureShare` parameters to the model for clipping
            self.protocol.algorithm.model.param_clipping = {  # type: ignore[attr-defined] # Reason: see above # noqa: B950
                "prime_q": self.protocol.aggregator._secure_share.prime_q,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
                "precision": self.protocol.aggregator._secure_share.precision,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
                "num_workers": len(pod_identifiers),
            }
        result = asyncio.run(self.run_async(pod_identifiers, require_all_pods))
        if model_out and result is not False:
            self._serialize(model_out)
        return result



"""Test Authorisation Checkers."""
import asyncio
from datetime import datetime, timedelta, timezone
import functools
import logging
import re
from types import MethodType
from typing import Callable, Dict, List, Literal, Optional, Type, Union, cast
from unittest.mock import AsyncMock, Mock, NonCallableMock, call, create_autospec

from _pytest.logging import LogCaptureFixture
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
from requests import HTTPError, RequestException
import responses

from bitfount import web_utils
from bitfount.federated.authorisation_checkers import (
    IdentityVerificationMethod,
    _LocalAuthorisation,
    _OIDCAuthorisationCode,
    _OIDCDeviceCode,
    _SAMLAuthorisation,
    _SignatureBasedAuthorisation,
    check_identity_verification_method,
)
from bitfount.federated.pod_response_message import _PodResponseMessage
from bitfount.federated.task_requests import (
    _EncryptedTaskRequest,
    _SignedEncryptedTaskRequest,
    _TaskRequest,
    _TaskRequestMessage,
)
from bitfount.federated.transport.types import (
    _OIDCAuthFlowResponse,
    _PodDeviceCodeDetails,
)
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.federated.types import (
    SerializedAlgorithm,
    SerializedProtocol,
    _PodResponseType,
    _TaskRequestMessageGenerator,
)
from bitfount.hub.api import BitfountAM, BitfountHub
from bitfount.hub.authentication_flow import (
    _AUTHORIZATION_PENDING_ERROR,
    _DEVICE_CODE_GRANT_TYPE,
    _SLOW_DOWN_ERROR,
    _AuthEnv,
)
from bitfount.hub.types import (
    _DeviceAccessTokenRequestDict,
    _DeviceAccessTokenResponseJSON,
    _PKCEAccessTokenResponseJSON,
)
from tests.utils import PytestRequest
from tests.utils.helper import get_info_logs, get_warning_logs, unit_test


@unit_test
class TestIdentityVerificationMethodChecker:
    """Tests for check_identity_verification_method()."""

    def test_check_identity_verification_method_passes_for_listed_methods(self) -> None:
        """Check function passes for expected values (str or enum)."""
        # Test works with string-based values
        for method in IdentityVerificationMethod:
            assert check_identity_verification_method(method.value) == method

        # Test works with enum-based values
        for method in IdentityVerificationMethod:
            assert check_identity_verification_method(method) == method

    def test_check_identity_verification_method_raises_exception_on_unrecognised(
        self,
    ) -> None:
        """Check function raises exception when passed unrecognised method."""
        fake_method = "fake_method"
        with pytest.raises(
            ValueError, match=f"Unsupported identity verification method: {fake_method}"
        ):
            check_identity_verification_method(fake_method)


@fixture
def pod_identifier() -> str:
    """Single pod identifier."""
    return "someUser/somePod"


@fixture
def pod_identifiers() -> List[str]:
    """List of pod identifiers."""
    return ["pod_1/identifier_1", "pod_1/identifier_2"]


@fixture
def serialized_protocol() -> SerializedProtocol:
    """Protocol details."""
    return SerializedProtocol(
        class_name="some protocol",
        algorithm=SerializedAlgorithm(class_name="some algorithm"),
    )


@fixture
def modeller_name() -> str:
    """Name of modeller."""
    return "someModeller"


@fixture(params=[True, False], ids=lambda arg: f"serialize_input={arg}")
def serialize_as_bytes(request: PytestRequest) -> bool:
    """Determines whether the method input should be serialized or not."""
    return cast(bool, request.param)


@fixture
def aes_key() -> bytes:
    """Fake AES key."""
    return b"an_aes_key"


@fixture
def mock_pod_public_key() -> NonCallableMock:
    """Mock pod RSA public key."""
    mock_pod_public_key: NonCallableMock = create_autospec(RSAPublicKey, instance=True)
    return mock_pod_public_key


@fixture
def mock_pod_private_key() -> NonCallableMock:
    """Mock pod RSA private key."""
    mock_pod_private_key: NonCallableMock = create_autospec(
        RSAPrivateKey, instance=True
    )
    return mock_pod_private_key


@fixture
def mock_task_request() -> NonCallableMock:
    """Mock TaskRequest instance.

    serialize() returns a static value.
    """
    mock_task_request: NonCallableMock = create_autospec(_TaskRequest, instance=True)
    mock_task_request.serialize.return_value = b"mock_task_request"
    return mock_task_request


@fixture
def mock_task_request_import(
    mock_task_request: NonCallableMock, mocker: MockerFixture
) -> Mock:
    """Mocks the TaskRequest import in authorisation_checkers.py.

    Constructor and deserialize method will return the mock instance defined.
    """
    mock_task_request_import: Mock = mocker.patch(
        "bitfount.federated.authorisation_checkers._TaskRequest",
        return_value=mock_task_request,
    )
    mock_task_request_import.deserialize.return_value = mock_task_request
    return mock_task_request_import


@fixture
def mock_encrypted_task_request() -> NonCallableMock:
    """Mock EncryptedTaskRequest instance.

    serialize() returns a static value.
    """
    mock_encrypted_task_request: NonCallableMock = create_autospec(
        _EncryptedTaskRequest, instance=True
    )
    mock_encrypted_task_request.serialize.return_value = b"mock_encrypted_task_request"
    return mock_encrypted_task_request


@fixture
def mock_encrypted_tr_import(
    mock_encrypted_task_request: NonCallableMock, mocker: MockerFixture
) -> Mock:
    """Mocks the EncryptedTaskRequest import in authorisation_checkers.py.

    Constructor and deserialize method will return the mock instance defined.
    """
    mock_encrypted_tr_import: Mock = mocker.patch(
        "bitfount.federated.authorisation_checkers._EncryptedTaskRequest",
        return_value=mock_encrypted_task_request,
    )
    mock_encrypted_tr_import.deserialize.return_value = mock_encrypted_task_request
    return mock_encrypted_tr_import


@fixture
def mock_signed_encrypted_task_request() -> NonCallableMock:
    """Mock SignedEncryptedTaskRequest instance.

    serialize() returns a static value.
    """
    mock_encrypted_task_request: NonCallableMock = create_autospec(
        _SignedEncryptedTaskRequest, instance=True
    )
    mock_encrypted_task_request.serialize.return_value = (
        b"mock_signed_encrypted_task_request"
    )
    return mock_encrypted_task_request


@fixture
def mock_signed_encrypted_tr_import(
    mock_signed_encrypted_task_request: NonCallableMock, mocker: MockerFixture
) -> Mock:
    """Mocks the SignedEncryptedTaskRequest import in authorisation_checkers.py.

    Constructor and deserialize method will return the mock instance defined.
    """
    mock_signed_encrypted_tr_import: Mock = mocker.patch(
        "bitfount.federated.authorisation_checkers._SignedEncryptedTaskRequest",
        return_value=mock_signed_encrypted_task_request,
    )
    mock_signed_encrypted_tr_import.deserialize.return_value = (
        mock_signed_encrypted_task_request
    )
    return mock_signed_encrypted_tr_import


@fixture
def mock_task_request_message() -> NonCallableMock:
    """Mock TaskRequestMessage instance.

    serialize() returns a static value.
    """
    mock_task_request_message: NonCallableMock = create_autospec(
        _TaskRequestMessage, instance=True
    )
    mock_task_request_message.serialize.return_value = b"mock_task_request_message"
    return mock_task_request_message


@fixture
def mock_task_request_message_import(
    mock_task_request_message: NonCallableMock, mocker: MockerFixture
) -> Mock:
    """Mocks the TaskRequestMessage import in authorisation_checkers.py.

    Constructor and deserialize method will return the mock instance defined.
    """
    mock_task_request_message_import: Mock = mocker.patch(
        "bitfount.federated.authorisation_checkers._TaskRequestMessage",
        return_value=mock_task_request_message,
    )
    mock_task_request_message_import.deserialize.return_value = (
        mock_task_request_message
    )
    return mock_task_request_message_import


@fixture
def task_request(
    aes_key: bytes, pod_identifiers: List[str], serialized_protocol: SerializedProtocol
) -> _TaskRequest:
    """Constructed TaskRequest instance."""
    return _TaskRequest(serialized_protocol, pod_identifiers, aes_key)


@fixture
def encrypted_task_request(task_request: _TaskRequest) -> _EncryptedTaskRequest:
    """Constructed EncryptedTaskRequest instance.

    The `encrypted_request` parameter is not actually encrypted, but just serialized
    to bytes.
    """
    # Not encrypted contents
    return _EncryptedTaskRequest(encrypted_request=task_request.serialize())


@fixture
def signed_encrypted_task_request(
    task_request: _TaskRequest,
) -> _SignedEncryptedTaskRequest:
    """Constructed SignedEncryptedTaskRequest instance.

    The `encrypted_request` parameter is not actually encrypted, but just serialized
    to bytes. The signature is a static value.
    """
    # Not encrypted contents, fake signature
    return _SignedEncryptedTaskRequest(
        encrypted_request=task_request.serialize(), signature=b"fake_signature"
    )


@fixture
def get_task_request_message(
    encrypted_task_request: _EncryptedTaskRequest,
    serialized_protocol: SerializedProtocol,
) -> Callable[[str], _TaskRequestMessage]:
    """Return function that will create TaskRequestMessage instances from auth_type."""

    def _make_task_request_message(auth_type: str) -> _TaskRequestMessage:
        return _TaskRequestMessage(
            serialized_protocol=serialized_protocol,
            auth_type=auth_type,
            request=encrypted_task_request.serialize(),
        )

    return _make_task_request_message


@fixture
def get_signed_task_request_message(
    serialized_protocol: SerializedProtocol,
    signed_encrypted_task_request: _EncryptedTaskRequest,
) -> Callable[[str], _TaskRequestMessage]:
    """Return function that will create TaskRequestMessage instances from auth_type.

    Inner message type is a signed and encrypted task request.
    """

    def _make_task_request_message(auth_type: str) -> _TaskRequestMessage:
        return _TaskRequestMessage(
            serialized_protocol=serialized_protocol,
            auth_type=auth_type,
            request=signed_encrypted_task_request.serialize(),
        )

    return _make_task_request_message


@unit_test
class TestLocalAuthorisation:
    """Tests Local Authorisation."""

    async def test_local_authorisation_approves(
        self,
        modeller_name: str,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Tests local authorisation.

        This authoriser 'approves' all requests for local use
        it simply returns the PodResponseMessage given to it.
        """
        expected_pod_response = _PodResponseMessage(modeller_name, pod_identifier)
        authoriser = _LocalAuthorisation(expected_pod_response, serialized_protocol)

        result = await authoriser.check_authorisation()

        assert result == expected_pod_response
        assert result.messages == {}

    def test__generate_task_request_message(
        self,
        aes_key: bytes,
        mock_encrypted_task_request: NonCallableMock,
        mock_encrypted_tr_import: Mock,
        mock_pod_public_key: NonCallableMock,
        mock_rsa_encryption: Mock,
        mock_task_request: NonCallableMock,
        mock_task_request_import: Mock,
        mock_task_request_message: NonCallableMock,
        mock_task_request_message_import: Mock,
        pod_identifiers: List[str],
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Tests the _generate_task_request_message method for Local Authorisation."""
        trm = _LocalAuthorisation._generate_task_request_message(
            serialized_protocol=serialized_protocol,
            pod_identifiers=pod_identifiers,
            aes_key=aes_key,
            pod_public_key=mock_pod_public_key,
        )

        # Check TaskRequest constructed as expected
        mock_task_request_import.assert_called_once_with(
            serialized_protocol, pod_identifiers, aes_key
        )
        # Check EncryptedTaskRequest constructed as expected (mocked RSA encryption
        # means serialize return value will be passed)
        mock_encrypted_tr_import.assert_called_once_with(
            mock_task_request.serialize.return_value
        )
        # Check encryption done
        mock_rsa_encryption.assert_called_once_with(
            mock_task_request.serialize.return_value,
            mock_pod_public_key,
        )
        # Check TaskRequestMessage constructed as expected
        mock_task_request_message_import.assert_called_once_with(
            serialized_protocol,
            IdentityVerificationMethod.LOCAL.value,
            mock_encrypted_task_request.serialize.return_value,
        )
        # Check that output value is serialized TaskRequestMessage
        assert trm == mock_task_request_message.serialize.return_value

    def test_extract_from_task_request_message(
        self,
        encrypted_task_request: _EncryptedTaskRequest,
        get_task_request_message: Callable[[str], _TaskRequestMessage],
        serialize_as_bytes: bool,
    ) -> None:
        """Tests extract_from_task_request_message for Local Authorisation.

        Tests for both deserialized and serialized inputs.
        """
        task_request_message = get_task_request_message(
            IdentityVerificationMethod.LOCAL.value
        )

        if serialize_as_bytes:
            extracted = _LocalAuthorisation.extract_from_task_request_message(
                task_request_message.serialize()
            )
        else:
            extracted = _LocalAuthorisation.extract_from_task_request_message(
                task_request_message
            )

        assert extracted == encrypted_task_request

    def test_create_task_request_message_generator(self) -> None:
        """Checks create_task_request_message_generator returns right type."""
        assert isinstance(
            _LocalAuthorisation.create_task_request_message_generator(),
            _TaskRequestMessageGenerator,
        )

    def test_unpack_task_request(
        self,
        get_task_request_message: Callable[[str], _TaskRequestMessage],
        mock_pod_private_key: NonCallableMock,
        mock_rsa_decryption: Mock,
        serialize_as_bytes: bool,
        task_request: _TaskRequest,
    ) -> None:
        """Tests unpack_task_request for Local Authorisation.

        Tests for both deserialized and serialized inputs.
        """
        task_request_message = get_task_request_message(
            IdentityVerificationMethod.LOCAL.value
        )

        if serialize_as_bytes:
            unpacked = _LocalAuthorisation.unpack_task_request(
                task_request_message.serialize(), mock_pod_private_key
            )
        else:
            unpacked = _LocalAuthorisation.unpack_task_request(
                task_request_message, mock_pod_private_key
            )

        assert unpacked == task_request


@unit_test
class TestSignatureBasedAuthorisation:
    """Tests Signature (Private Key) Authorisation Checker."""

    @fixture
    def mock_hub(self) -> Mock:
        """Mock hub."""
        hub: Mock = create_autospec(BitfountHub, instance=True)
        return hub

    @pytest.mark.parametrize(
        "response_type,expected_messages",
        [
            (_PodResponseType.ACCEPT, {}),
            (
                _PodResponseType.UNAUTHORISED,
                {_PodResponseType.UNAUTHORISED.name: []},
            ),
        ],
    )
    async def test_signature_authorisation_returns_responses(
        self,
        expected_messages: Dict[str, str],
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        modeller_name: str,
        pod_identifier: str,
        response_type: _PodResponseType,
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Test Signature Authorisation returns expected PodResponseMessages."""
        encrypted_task = b"some task"
        signature = b"some signature"

        # Mock out API _check_access call
        mock_access_manager.check_signature_based_access_request.return_value = (
            response_type
        )

        authoriser = _SignatureBasedAuthorisation(
            pod_response_message=_PodResponseMessage(modeller_name, pod_identifier),
            access_manager=mock_access_manager,
            modeller_name=mock_worker_mailbox.modeller_name,
            encrypted_task_request=encrypted_task,
            signature=signature,
            serialized_protocol=serialized_protocol,
        )

        result = await authoriser.check_authorisation()

        # Check API call made correctly
        mock_access_manager.check_signature_based_access_request.assert_called_once_with(
            unsigned_task=encrypted_task,
            signed_task=_SignedEncryptedTaskRequest(
                encrypted_task, signature
            ).serialize(),
            pod_identifier=pod_identifier,
            serialized_protocol=serialized_protocol,
            modeller_name=modeller_name,
        )

        assert result.messages == expected_messages

    def test__generate_task_request_message(
        self,
        aes_key: bytes,
        mock_pod_private_key: NonCallableMock,
        mock_pod_public_key: NonCallableMock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
        mock_signed_encrypted_task_request: NonCallableMock,
        mock_signed_encrypted_tr_import: Mock,
        mock_task_request: NonCallableMock,
        mock_task_request_import: Mock,
        mock_task_request_message: NonCallableMock,
        mock_task_request_message_import: Mock,
        pod_identifiers: List[str],
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Test the generate_task_request_message method in sig-based authorisation."""
        # Create inner message maker class
        message_maker = _SignatureBasedAuthorisation._SignedMessageMaker(
            mock_pod_private_key
        )

        # Call
        trm = message_maker.generate_task_request_message(
            serialized_protocol=serialized_protocol,
            pod_identifiers=pod_identifiers,
            aes_key=aes_key,
            pod_public_key=mock_pod_public_key,
        )

        # Check TaskRequest constructed as expected
        mock_task_request_import.assert_called_once_with(
            serialized_protocol, pod_identifiers, aes_key
        )
        # Check encryption done
        mock_rsa_encryption.assert_called_once_with(
            mock_task_request.serialize.return_value,
            mock_pod_public_key,
        )
        # Check signing done (mocked RSA encryption means serialize() return value
        # will be passed)
        mock_rsa_sign_message.assert_called_once_with(
            mock_pod_private_key, mock_task_request.serialize.return_value
        )
        # Check SignedEncryptedTaskRequest constructed as expected (mocked RSA
        # encryption means serialize() return value will be passed; mocked signing
        # means static value passed)
        mock_signed_encrypted_tr_import.assert_called_once_with(
            mock_task_request.serialize.return_value,
            mock_rsa_sign_message(None, None),  # extract signature value
        )
        # Check TaskRequestMessage constructed as expected
        mock_task_request_message_import.assert_called_once_with(
            serialized_protocol,
            IdentityVerificationMethod.KEYS.value,
            mock_signed_encrypted_task_request.serialize.return_value,
        )
        # Check that output value is serialized TaskRequestMessage
        assert trm == mock_task_request_message.serialize.return_value

    def test_create_task_request_message_generator(
        self,
        mock_pod_private_key: NonCallableMock,
    ) -> None:
        """Test _SignatureBasedAuthorisation.create_task_request_message_generator()."""
        message_generator = (
            _SignatureBasedAuthorisation.create_task_request_message_generator(
                mock_pod_private_key
            )
        )

        # Check right type
        assert isinstance(message_generator, _TaskRequestMessageGenerator)
        # Check key matches in created inner class
        wrapper_class = cast(MethodType, message_generator).__self__
        assert isinstance(
            wrapper_class, _SignatureBasedAuthorisation._SignedMessageMaker
        )
        assert wrapper_class.private_key == mock_pod_private_key

    def test_extract_from_task_request_message(
        self,
        get_signed_task_request_message: Callable[[str], _TaskRequestMessage],
        serialize_as_bytes: bool,
        signed_encrypted_task_request: _SignedEncryptedTaskRequest,
    ) -> None:
        """Test _SignatureBasedAuthorisation.extract_from_task_request_message().

        Tests for both deserialized and serialized inputs.
        """
        task_request_message = get_signed_task_request_message(
            IdentityVerificationMethod.KEYS.value
        )

        if serialize_as_bytes:
            extracted = _SignatureBasedAuthorisation.extract_from_task_request_message(
                task_request_message.serialize()
            )
        else:
            extracted = _SignatureBasedAuthorisation.extract_from_task_request_message(
                task_request_message
            )

        assert extracted == signed_encrypted_task_request

    def test_unpack_task_request(
        self,
        get_signed_task_request_message: Callable[[str], _TaskRequestMessage],
        mock_pod_private_key: NonCallableMock,
        mock_rsa_decryption: Mock,
        serialize_as_bytes: bool,
        task_request: _TaskRequest,
    ) -> None:
        """Test _SignatureBasedAuthorisation.unpack_task_request().

        Tests for both deserialized and serialized inputs.
        """
        task_request_message = get_signed_task_request_message(
            IdentityVerificationMethod.KEYS.value
        )

        if serialize_as_bytes:
            unpacked = _SignatureBasedAuthorisation.unpack_task_request(
                task_request_message.serialize(), mock_pod_private_key
            )
        else:
            unpacked = _SignatureBasedAuthorisation.unpack_task_request(
                task_request_message, mock_pod_private_key
            )

        assert unpacked == task_request


@unit_test
class TestSAMLAuthorisation:
    """Tests SAML Authorisation Checker."""

    @fixture
    def pod_name(self) -> str:
        """Pod name."""
        return "somePod"

    @fixture
    def saml_request(self) -> str:
        """SAML Request."""
        return "some saml request"

    @fixture
    def saml_id(self) -> str:
        """SAML Request ID."""
        return "some saml ID"

    @fixture
    def saml_response(self) -> str:
        """SAML Response."""
        return "some saml response"

    @fixture
    def access_manager(self, saml_id: str, saml_request: str) -> Mock:
        """Mock access manager."""
        access_manager: Mock = create_autospec(BitfountAM, instance=True)
        access_manager.get_saml_challenge.return_value = saml_request, saml_id
        return access_manager

    @fixture
    def worker_mailbox(
        self, modeller_name: str, pod_identifier: str, saml_response: str
    ) -> Mock:
        """Mock worker mailbox."""
        mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
        mailbox.get_saml_response = AsyncMock(return_value=saml_response)
        mailbox.modeller_name = modeller_name
        mailbox.pod_identifier = pod_identifier
        return mailbox

    @pytest.mark.parametrize(
        "response_type,expected_messages",
        [
            (_PodResponseType.ACCEPT, {}),
            (
                _PodResponseType.UNAUTHORISED,
                {_PodResponseType.UNAUTHORISED.name: []},
            ),
        ],
    )
    async def test_saml_authorisation_returns_responses(
        self,
        access_manager: Mock,
        expected_messages: Dict[str, List[str]],
        modeller_name: str,
        pod_identifier: str,
        pod_name: str,
        response_type: _PodResponseType,
        saml_id: str,
        saml_request: str,
        saml_response: str,
        serialized_protocol: SerializedProtocol,
        worker_mailbox: Mock,
    ) -> None:
        """Test SAML Authorisation returns expected PodResponseMessages."""
        access_manager.validate_saml_response.return_value = response_type
        response_message = _PodResponseMessage(modeller_name, pod_identifier)
        authoriser = _SAMLAuthorisation(
            pod_response_message=response_message,
            access_manager=access_manager,
            mailbox=worker_mailbox,
            serialized_protocol=serialized_protocol,
        )

        result = await authoriser.check_authorisation()

        worker_mailbox.issue_saml_challenge.assert_called_once_with(saml_request)
        access_manager.validate_saml_response.assert_called_once_with(
            saml_response, saml_id, pod_identifier, modeller_name, serialized_protocol
        )

        assert result.messages == expected_messages

    def test_extract_from_task_request_message(
        self,
        encrypted_task_request: _EncryptedTaskRequest,
        get_task_request_message: Callable[[str], _TaskRequestMessage],
        serialize_as_bytes: bool,
    ) -> None:
        """Tests extract_from_task_request_message for SAML Authorisation.

        Tests for both deserialized and serialized inputs.
        """
        task_request_message = get_task_request_message(
            IdentityVerificationMethod.SAML.value
        )

        if serialize_as_bytes:
            extracted = _SAMLAuthorisation.extract_from_task_request_message(
                task_request_message.serialize()
            )
        else:
            extracted = _SAMLAuthorisation.extract_from_task_request_message(
                task_request_message
            )

        assert extracted == encrypted_task_request


@fixture
def mock_access_manager() -> Mock:
    """Mock access manager."""
    access_manager: Mock = create_autospec(BitfountAM, instance=True)
    return access_manager


@fixture
def mock_worker_mailbox(modeller_name: str, pod_identifier: str) -> Mock:
    """Mock worker mailbox."""
    mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
    mailbox.modeller_name = modeller_name
    mailbox.pod_identifier = pod_identifier
    return mailbox


@fixture
def fake_auth_domain() -> str:
    """A fake auth domain for tests."""
    return "fake_auth_domain"


@fixture
def fake_client_id() -> str:
    """A fake client ID for tests."""
    return "fake_client_id"


@unit_test
class TestOIDCAuthCodeAuthorisation:
    """Tests for _OIDCAuthorisationCode authoriser."""

    @fixture
    def auth_code(self) -> str:
        """Auth code for tests."""
        return "auth_code_for_tests"

    @fixture
    def code_verifier(self) -> str:
        """Code verifier for tests."""
        return "code_verifier_for_tests"

    @fixture
    def redirect_uri(self) -> str:
        """Redirect URI for tests."""
        return "redirect_uri"

    @fixture
    def pkce_access_token_response_json(self) -> _PKCEAccessTokenResponseJSON:
        """Response JSON from /oauth/token."""
        response_json: _PKCEAccessTokenResponseJSON = {
            "access_token": "access_token_value",
            "refresh_token": "refresh_token_value",
            "id_token": "id_token_value",
            "token_type": "Bearer",
            "expires_in": 60 * 60,
        }
        return response_json

    @fixture
    def oidc_auth_code_authoriser(
        self,
        fake_auth_domain: str,
        fake_client_id: str,
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        modeller_name: str,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
    ) -> _OIDCAuthorisationCode:
        """Constructed OIDC authoriser for tests."""
        pod_response_message = _PodResponseMessage(modeller_name, pod_identifier)
        return _OIDCAuthorisationCode(
            pod_response_message=pod_response_message,
            access_manager=mock_access_manager,
            mailbox=mock_worker_mailbox,
            serialized_protocol=serialized_protocol,
            _auth_domain=fake_auth_domain,
            _client_id=fake_client_id,
        )

    def test__generate_task_request_message(
        self,
        aes_key: bytes,
        mock_encrypted_task_request: NonCallableMock,
        mock_encrypted_tr_import: Mock,
        mock_pod_public_key: NonCallableMock,
        mock_rsa_encryption: Mock,
        mock_task_request: NonCallableMock,
        mock_task_request_import: Mock,
        mock_task_request_message: NonCallableMock,
        mock_task_request_message_import: Mock,
        pod_identifiers: List[str],
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Test _generate_task_request_message for OIDC Authorisation."""
        trm = _OIDCAuthorisationCode._generate_task_request_message(
            serialized_protocol=serialized_protocol,
            pod_identifiers=pod_identifiers,
            aes_key=aes_key,
            pod_public_key=mock_pod_public_key,
        )

        # Check TaskRequest constructed as expected
        mock_task_request_import.assert_called_once_with(
            serialized_protocol, pod_identifiers, aes_key
        )
        # Check EncryptedTaskRequest constructed as expected (mocked RSA encryption
        # means serialize return value will be passed)
        mock_encrypted_tr_import.assert_called_once_with(
            mock_task_request.serialize.return_value
        )
        # Check encryption done
        mock_rsa_encryption.assert_called_once_with(
            mock_task_request.serialize.return_value,
            mock_pod_public_key,
        )
        # Check TaskRequestMessage constructed as expected
        mock_task_request_message_import.assert_called_once_with(
            serialized_protocol,
            IdentityVerificationMethod.OIDC_ACF_PKCE.value,
            mock_encrypted_task_request.serialize.return_value,
        )
        # Check that output value is serialized TaskRequestMessage
        assert trm == mock_task_request_message.serialize.return_value

    def test_create_task_request_message_generator(self) -> None:
        """Tests return of OIDC create_task_request_message_generator()."""
        assert isinstance(
            _OIDCAuthorisationCode.create_task_request_message_generator(),
            _TaskRequestMessageGenerator,
        )

    def test__get_client_id(
        self, fake_client_id: str, oidc_auth_code_authoriser: _OIDCAuthorisationCode
    ) -> None:
        """Tests client ID retrieval."""
        # Currently, we expect this to return a static value
        assert oidc_auth_code_authoriser._get_client_id() == fake_client_id

    @responses.activate
    def test__get_access_token(
        self,
        auth_code: str,
        code_verifier: str,
        oidc_auth_code_authoriser: _OIDCAuthorisationCode,
        pkce_access_token_response_json: _PKCEAccessTokenResponseJSON,
        redirect_uri: str,
    ) -> None:
        """Test the _get_access_token POST request and return value."""
        responses.add(
            method=responses.POST,
            url=oidc_auth_code_authoriser._token_endpoint,
            json=pkce_access_token_response_json,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "authorization_code",
                        "client_id": oidc_auth_code_authoriser._client_id,
                        "code": auth_code,
                        "code_verifier": code_verifier,
                        "redirect_uri": redirect_uri,
                    }
                )
            ],
        )

        received_access_token = oidc_auth_code_authoriser._get_access_token(
            auth_code, code_verifier, redirect_uri
        )

        assert received_access_token == pkce_access_token_response_json["access_token"]

    @responses.activate
    def test__get_access_token_non_200_response(
        self,
        auth_code: str,
        code_verifier: str,
        oidc_auth_code_authoriser: _OIDCAuthorisationCode,
        redirect_uri: str,
    ) -> None:
        """Tests _get_access_token throws error, non-200 response."""
        responses.add(
            method=responses.POST,
            url=oidc_auth_code_authoriser._token_endpoint,
            status=400,
        )

        with pytest.raises(HTTPError):
            oidc_auth_code_authoriser._get_access_token(
                auth_code, code_verifier, redirect_uri
            )

    @pytest.mark.parametrize(
        argnames="check_access_response",
        argvalues=(
            pytest.param(_PodResponseType.ACCEPT, id="accept"),
            pytest.param(_PodResponseType.NO_ACCESS, id="no_access"),
            pytest.param(
                _PodResponseType.INVALID_PROOF_OF_IDENTITY,
                id="invalid_proof_of_identity",
            ),
            pytest.param(_PodResponseType.UNAUTHORISED, id="unauthorised"),
            pytest.param(
                _PodResponseType.NO_PROOF_OF_IDENTITY, id="no_proof_of_identity"
            ),
        ),
    )
    async def test_check_authorisation(
        self,
        auth_code: str,
        caplog: LogCaptureFixture,
        check_access_response: _PodResponseType,
        code_verifier: str,
        fake_client_id: str,
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        mocker: MockerFixture,
        modeller_name: str,
        oidc_auth_code_authoriser: _OIDCAuthorisationCode,
        pkce_access_token_response_json: _PKCEAccessTokenResponseJSON,
        pod_identifier: str,
        redirect_uri: str,
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Test authorisation checking for OIDC flow."""
        # Mock out received values from modeller
        mock_worker_mailbox.get_oidc_auth_flow_response.return_value = (
            _OIDCAuthFlowResponse(
                auth_code,
                code_verifier,
                redirect_uri,
            )
        )

        # Mock out _get_access_token
        mock__get_access_token = mocker.patch.object(
            oidc_auth_code_authoriser,
            "_get_access_token",
            autospec=True,
            return_value=pkce_access_token_response_json["access_token"],
        )

        # Mock out API _check_access call
        mock_access_manager.check_oidc_access_request.return_value = (
            check_access_response
        )

        with caplog.at_level(logging.INFO):
            response_message = await oidc_auth_code_authoriser.check_authorisation()

        # Check correct Client ID sent
        mock_worker_mailbox.send_oidc_client_id.assert_awaited_once_with(fake_client_id)
        # Check _get_access_token called as expected
        mock__get_access_token.assert_called_once_with(
            auth_code, code_verifier, redirect_uri
        )
        # Check API call made correctly
        mock_access_manager.check_oidc_access_request.assert_called_once_with(
            pod_identifier=pod_identifier,
            serialized_protocol=serialized_protocol,
            modeller_name=modeller_name,
            modeller_access_token=pkce_access_token_response_json["access_token"],
        )
        # Check logs
        info_logs = get_info_logs(caplog)
        assert (
            f"OIDC authorisation check complete. "
            f"Will inform modeller: {check_access_response.name}" in info_logs
        )

        # Check return
        expected_pod_response = _PodResponseMessage(modeller_name, pod_identifier)
        expected_pod_response.add(check_access_response)
        assert response_message == expected_pod_response


@unit_test
class TestOIDCDeviceCodeAuthorisation:
    """Tests for _OIDCDeviceCode authoriser."""

    @fixture
    def oidc_device_code_authoriser(
        self,
        fake_auth_domain: str,
        fake_client_id: str,
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        modeller_name: str,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
    ) -> _OIDCDeviceCode:
        """An _OIDCDeviceCode authoriser instance."""
        pod_response_message = _PodResponseMessage(modeller_name, pod_identifier)
        return _OIDCDeviceCode(
            pod_response_message=pod_response_message,
            access_manager=mock_access_manager,
            mailbox=mock_worker_mailbox,
            serialized_protocol=serialized_protocol,
            _auth_domain=fake_auth_domain,
            _client_id=fake_client_id,
        )

    def test_auth_domain_and_client_id_always_set(
        self,
        fake_auth_domain: str,
        fake_client_id: str,
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        mocker: MockerFixture,
        modeller_name: str,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Tests default values used if auth_domain and client_id are not set."""
        # Mock _get_auth_environment() function
        mocker.patch(
            "bitfount.federated.authorisation_checkers._get_auth_environment",
            autospec=True,
            return_value=_AuthEnv(
                name="auth_env_name",
                auth_domain=fake_auth_domain,
                client_id=fake_client_id,
            ),
        )

        # Create _OIDCDeviceCode without providing auth_domain or client_id
        pod_response_message = _PodResponseMessage(modeller_name, pod_identifier)
        oidc_device_code_authoriser = _OIDCDeviceCode(
            pod_response_message=pod_response_message,
            access_manager=mock_access_manager,
            mailbox=mock_worker_mailbox,
            serialized_protocol=serialized_protocol,
        )

        # Check defaults set
        assert oidc_device_code_authoriser._auth_domain == fake_auth_domain
        assert oidc_device_code_authoriser._client_id == fake_client_id

    def test__generate_task_request_message(
        self,
        aes_key: bytes,
        mock_encrypted_task_request: NonCallableMock,
        mock_encrypted_tr_import: Mock,
        mock_pod_public_key: NonCallableMock,
        mock_rsa_encryption: Mock,
        mock_task_request: NonCallableMock,
        mock_task_request_import: Mock,
        mock_task_request_message: NonCallableMock,
        mock_task_request_message_import: Mock,
        pod_identifiers: List[str],
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Test _generate_task_request_message for OIDC Authorisation."""
        trm = _OIDCDeviceCode._generate_task_request_message(
            serialized_protocol=serialized_protocol,
            pod_identifiers=pod_identifiers,
            aes_key=aes_key,
            pod_public_key=mock_pod_public_key,
        )

        # Check TaskRequest constructed as expected
        mock_task_request_import.assert_called_once_with(
            serialized_protocol, pod_identifiers, aes_key
        )
        # Check EncryptedTaskRequest constructed as expected (mocked RSA encryption
        # means serialize return value will be passed)
        mock_encrypted_tr_import.assert_called_once_with(
            mock_task_request.serialize.return_value
        )
        # Check encryption done
        mock_rsa_encryption.assert_called_once_with(
            mock_task_request.serialize.return_value,
            mock_pod_public_key,
        )
        # Check TaskRequestMessage constructed as expected
        mock_task_request_message_import.assert_called_once_with(
            serialized_protocol,
            IdentityVerificationMethod.OIDC_DEVICE_CODE.value,
            mock_encrypted_task_request.serialize.return_value,
        )
        # Check that output value is serialized TaskRequestMessage
        assert trm == mock_task_request_message.serialize.return_value

    def test_create_task_request_message_generator(self) -> None:
        """Tests return of OIDC create_task_request_message_generator()."""
        assert isinstance(
            _OIDCDeviceCode.create_task_request_message_generator(),
            _TaskRequestMessageGenerator,
        )

    @fixture
    def device_code(self) -> str:
        """Device code."""
        return "someDeviceCode"

    @fixture
    def expires_at(self) -> datetime:
        """The code expiration time as a datetime."""
        return datetime.now(timezone.utc) + timedelta(seconds=900)

    @fixture
    def interval(self) -> int:
        """The polling interval to use with the device code."""
        return 5

    @fixture
    def token_endpoint(self, fake_auth_domain: str) -> str:
        """The expected token API endpoint."""
        return f"https://{fake_auth_domain}/oauth/token"

    @fixture
    def token_request(
        self, device_code: str, fake_client_id: str
    ) -> _DeviceAccessTokenRequestDict:
        """Expected request data for /oauth/token."""
        return {
            "client_id": fake_client_id,
            "grant_type": _DEVICE_CODE_GRANT_TYPE,
            "device_code": device_code,
        }

    @fixture
    def access_token(self) -> str:
        """Access token."""
        return "someAccessToken"

    @fixture
    def token_response(self, access_token: str) -> _DeviceAccessTokenResponseJSON:
        """Expected response data from /oauth/token."""
        return {
            "access_token": access_token,
            "id_token": "eyJ...0NE",
            "refresh_token": "eyJ...MoQ",
            "scope": "...",
            "expires_in": 86400,
            "token_type": "Bearer",
        }

    @fixture
    def mock_asyncio_sleep(self, mocker: MockerFixture) -> Mock:
        """Mock out asyncio.sleep() calls."""
        return mocker.patch.object(asyncio, "sleep", autospec=True)

    async def test__poll_for_access_token(
        self,
        access_token: str,
        caplog: LogCaptureFixture,
        device_code: str,
        expires_at: datetime,
        interval: int,
        mock_asyncio_sleep: Mock,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        token_endpoint: str,
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Test _poll_for_access_token() works correctly."""
        # We have to use the context manager as decorator doesn't work on async.
        # Can change when: https://github.com/getsentry/responses/pull/478 is released.
        with responses.RequestsMock() as rsps:
            # Add one initial "not ready yet" response
            rsps.add(
                responses.POST,
                token_endpoint,
                status=400,
                json={"error": _AUTHORIZATION_PENDING_ERROR},
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )
            # Then next response is the passing one
            rsps.add(
                responses.POST,
                token_endpoint,
                json=token_response,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )

            with caplog.at_level(logging.INFO):
                received_access_token = (
                    await oidc_device_code_authoriser._poll_for_access_token(
                        device_code, expires_at, interval
                    )
                )

            # Check received info
            assert received_access_token == access_token

            # Check slept once
            mock_asyncio_sleep.assert_awaited_once_with(interval)
            # Check logged out that still waiting
            info_logs = get_info_logs(caplog)
            assert (
                "Waiting on modeller to approve identity verification request..."
                in info_logs
            )

    async def test__poll_for_access_token_handles_slow_down_responses(
        self,
        access_token: str,
        caplog: LogCaptureFixture,
        device_code: str,
        expires_at: datetime,
        interval: int,
        mock_asyncio_sleep: Mock,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        token_endpoint: str,
        token_request: _DeviceAccessTokenRequestDict,
        token_response: _DeviceAccessTokenResponseJSON,
    ) -> None:
        """Tests _poll_for_access_token() handles slow down responses."""
        # We have to use the context manager as decorator doesn't work on async.
        # Can change when: https://github.com/getsentry/responses/pull/478 is released.
        with responses.RequestsMock() as rsps:
            # Add two initial "slow down" responses
            for _ in range(2):
                rsps.add(
                    responses.POST,
                    token_endpoint,
                    status=400,
                    json={"error": _SLOW_DOWN_ERROR},
                    match=[
                        responses.matchers.urlencoded_params_matcher(
                            cast(Dict[str, str], token_request)
                        )
                    ],
                )
            # Then next response is the passing one
            rsps.add(
                responses.POST,
                token_endpoint,
                json=token_response,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )

            received_access_token = (
                await oidc_device_code_authoriser._poll_for_access_token(
                    device_code, expires_at, interval
                )
            )

            # Check received info
            assert received_access_token == access_token

            # Check slept for expected times and told user
            mock_asyncio_sleep.assert_has_awaits(
                [call(interval + 1), call(interval + 2)]
            )
            warning_logs = get_warning_logs(caplog)
            assert (
                f"Polling token endpoint too frequently. "
                f"Increasing interval from {interval} to {interval+1}" in warning_logs
            )
            assert (
                f"Polling token endpoint too frequently. "
                f"Increasing interval from {interval+1} to {interval + 2}"
                in warning_logs
            )

    @pytest.mark.parametrize(
        argnames=("content_type", "content_value", "expected_error_msg"),
        argvalues=(
            pytest.param(
                "json",
                {},
                'Success response, but JSON was invalid: "{}"',
                id="bad JSON response",
            ),
            pytest.param(
                "body",
                "not a json response",
                'Success response, but JSON was invalid: "not a json response"',
                id="not JSON response",
            ),
        ),
    )
    async def test__poll_for_access_token_fails_bad_200(
        self,
        content_type: Literal["json", "body"],
        content_value: Union[dict, str],
        device_code: str,
        expected_error_msg: str,
        expires_at: datetime,
        interval: int,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        token_endpoint: str,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """Tests _poll_for_access_token() handles malformed 200 responses."""
        # We have to use the context manager as decorator doesn't work on async.
        # Can change when: https://github.com/getsentry/responses/pull/478 is released.
        with responses.RequestsMock() as rsps:
            partial_response_add = functools.partial(
                rsps.add,
                responses.POST,
                token_endpoint,
                status=200,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )
            if content_type == "json":
                partial_response_add(json=content_value)
            else:
                partial_response_add(body=content_value)

            with pytest.raises(
                RequestException,
                match=re.escape(expected_error_msg),
            ):
                await oidc_device_code_authoriser._poll_for_access_token(
                    device_code, expires_at, interval
                )

    @pytest.mark.parametrize(
        argnames=("content_type", "content_value", "expected_error_msg"),
        argvalues=(
            pytest.param(
                "json",
                {"error": "unexpected error"},
                "Error in OAuth response (400) (unexpected error)",
                id="non retry error",
            ),
            pytest.param(
                "json",
                {
                    "error": "unexpected error",
                    "error_description": "unexpected error description",
                },
                (
                    "Error in OAuth response (400) (unexpected error): "
                    "unexpected error description"
                ),
                id="non retry error with description",
            ),
            pytest.param(
                "body",
                "not a json response",
                '400 response, but JSON was invalid: "not a json response"',
                id="not JSON response",
            ),
        ),
    )
    async def test__poll_for_access_token_fails_bad_400(
        self,
        content_type: Literal["json", "body"],
        content_value: Union[dict, str],
        device_code: str,
        expected_error_msg: str,
        expires_at: datetime,
        interval: int,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        token_endpoint: str,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """Tests _poll_for_access_token() handles malformed 400 responses."""
        # We have to use the context manager as decorator doesn't work on async.
        # Can change when: https://github.com/getsentry/responses/pull/478 is released.
        with responses.RequestsMock() as rsps:
            partial_response_add = functools.partial(
                rsps.add,
                responses.POST,
                token_endpoint,
                status=400,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )
            if content_type == "json":
                partial_response_add(json=content_value)
            else:
                partial_response_add(body=content_value)

            with pytest.raises(
                RequestException,
                match=re.escape(expected_error_msg),
            ):
                await oidc_device_code_authoriser._poll_for_access_token(
                    device_code, expires_at, interval
                )

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=(
            100,
            201,
            300,
            401,
            500,
        ),
    )
    @pytest.mark.parametrize(
        argnames=(
            "content_type",
            "content_value",
            "expected_error_type",
            "expected_error_msg",
        ),
        argvalues=(
            pytest.param(
                "json",
                {},
                HTTPError,
                r'An unexpected error occurred: status code: [0-9]{3}; "{}"',
                id="empty JSON",
            ),
            pytest.param(
                "json",
                {"error": "error name"},
                HTTPError,
                r"Error in OAuth response \([0-9]{3}\) \(error name\)",
                id="error JSON",
            ),
            pytest.param(
                "json",
                {"error": "error name", "error_description": "error description"},
                HTTPError,
                (
                    r"Error in OAuth response \([0-9]{3}\) \(error name\): "
                    r"error description"
                ),
                id="error with description JSON",
            ),
            pytest.param(
                "body",
                "not a json response",
                RequestException,
                r'[0-9]{3} response, but JSON was invalid: "not a json response"',
                id="non-JSON response",
            ),
        ),
    )
    async def test__poll_for_access_token_fails_non_200_non_400(
        self,
        content_type: Literal["json", "body"],
        content_value: Union[dict, str],
        device_code: str,
        expected_error_msg: str,
        expected_error_type: Type[Exception],
        expires_at: datetime,
        interval: int,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        token_endpoint: str,
        token_request: _DeviceAccessTokenRequestDict,
    ) -> None:
        """Tests _poll_for_access_token() fails if non-200/400 response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        # We have to use the context manager as decorator doesn't work on async.
        # Can change when: https://github.com/getsentry/responses/pull/478 is released.
        with responses.RequestsMock() as rsps:
            partial_response_add = functools.partial(
                rsps.add,
                responses.POST,
                token_endpoint,
                status=status_code,
                match=[
                    responses.matchers.urlencoded_params_matcher(
                        cast(Dict[str, str], token_request)
                    )
                ],
            )
            if content_type == "json":
                partial_response_add(json=content_value)
            else:
                partial_response_add(body=content_value)

            with pytest.raises(
                expected_error_type,
                match=expected_error_msg,
            ):
                await oidc_device_code_authoriser._poll_for_access_token(
                    device_code, expires_at, interval
                )

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    async def test__poll_for_access_token_fails_when_expired(
        self,
        device_code: str,
        expires_at: datetime,
        interval: int,
        mocker: MockerFixture,
        oidc_device_code_authoriser: _OIDCDeviceCode,
    ) -> None:
        """Tests _poll_for_access_token() fails when device code expires."""
        # Patch out expiration check
        mocker.patch.object(
            oidc_device_code_authoriser, "_has_not_expired", return_value=False
        )

        with pytest.raises(
            TimeoutError,
            match=re.escape(
                "Device code has expired, unable to retrieve access token."
            ),
        ):
            await oidc_device_code_authoriser._poll_for_access_token(
                device_code, expires_at, interval
            )

    @pytest.mark.parametrize(
        argnames="check_access_response",
        argvalues=(
            pytest.param(_PodResponseType.ACCEPT, id="accept"),
            pytest.param(_PodResponseType.NO_ACCESS, id="no_access"),
            pytest.param(
                _PodResponseType.INVALID_PROOF_OF_IDENTITY,
                id="invalid_proof_of_identity",
            ),
            pytest.param(_PodResponseType.UNAUTHORISED, id="unauthorised"),
            pytest.param(
                _PodResponseType.NO_PROOF_OF_IDENTITY, id="no_proof_of_identity"
            ),
        ),
    )
    async def test_check_authorisation(
        self,
        access_token: str,
        caplog: LogCaptureFixture,
        check_access_response: _PodResponseType,
        device_code: str,
        expires_at: datetime,
        fake_client_id: str,
        interval: int,
        mock_access_manager: Mock,
        mock_worker_mailbox: Mock,
        mocker: MockerFixture,
        modeller_name: str,
        oidc_device_code_authoriser: _OIDCDeviceCode,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
    ) -> None:
        """Tests check_authorisation() works as expected."""
        # Patch out receiving device code details
        mock_worker_mailbox.get_oidc_device_code_response.return_value = (
            _PodDeviceCodeDetails(
                device_code=device_code,
                expires_at=expires_at,
                interval=interval,
            )
        )

        # Patch out polling for access token
        mock_poll_for_access_token = mocker.patch.object(
            oidc_device_code_authoriser,
            "_poll_for_access_token",
            autospec=True,
            return_value=access_token,
        )

        # Mock out API _check_access call
        mock_access_manager.check_oidc_access_request.return_value = (
            check_access_response
        )

        # Call
        with caplog.at_level(logging.INFO):
            response_message = await oidc_device_code_authoriser.check_authorisation()

        # Check polling for access token occurred
        mock_poll_for_access_token.assert_awaited_once_with(
            device_code, expires_at, interval
        )

        # Check API call made correctly
        mock_access_manager.check_oidc_access_request.assert_called_once_with(
            pod_identifier=pod_identifier,
            serialized_protocol=serialized_protocol,
            modeller_name=modeller_name,
            modeller_access_token=access_token,
        )

        # Check logs
        info_logs = get_info_logs(caplog)
        assert (
            f"OIDC authorisation check complete. "
            f"Will inform modeller: {check_access_response.name}" in info_logs
        )

        # Check return
        expected_pod_response = _PodResponseMessage(modeller_name, pod_identifier)
        expected_pod_response.add(check_access_response)
        assert response_message == expected_pod_response



"""Models using LightGBM as the backend."""
from __future__ import annotations

from typing import Any

from bitfount.backends.lightgbm.models.base_models import BaseLGBMRandomForest
from bitfount.models.base_models import ClassifierMixIn, RegressorMixIn
from bitfount.utils import delegates


@delegates()
class LGBMRandomForestClassifier(ClassifierMixIn, BaseLGBMRandomForest):
    """LGBM Classifier for Random Forests and GBMs.

    Currently only supports binary classification.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._set_objective()
        self._set_training_metrics()

    def _set_objective(self) -> None:
        """Set training objective."""
        self.objective: str = "binary"

    def _set_training_metrics(self) -> None:
        """Set training metrics."""
        self.training_metrics = ["binary_logloss", "auc"]


@delegates()
class LGBMRandomForestRegressor(RegressorMixIn, BaseLGBMRandomForest):
    """LGBM Regressor for Random Forests and GBMs."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._set_objective()
        self._set_training_metrics()

    def _set_objective(self) -> None:
        """Set training objective."""
        self.objective = "regression"

    def _set_training_metrics(self) -> None:
        """Set training metrics."""
        self.training_metrics = ["mae", "mse", "rmse"]
